"""This file contains some examples of modules that can be used with GFN."""

import logging
import math
from abc import ABC, abstractmethod
from typing import Literal, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from linear_attention_transformer import LinearAttentionTransformer
from tensordict import TensorDict
from torch import Tensor
from torch_geometric.nn import DirGNNConv, GCNConv, GINConv

from gfn.actions import GraphActions, GraphActionType
from gfn.utils.common import is_int_dtype
from gfn.utils.graphs import GeometricBatch, get_edge_indices

logger = logging.getLogger(__name__)

ACTIVATION_FNS = {
    "relu": nn.ReLU,
    "leaky_relu": nn.LeakyReLU,
    "tanh": nn.Tanh,
    "elu": nn.ELU,
}


class MLP(nn.Module):
    """Implements a basic MLP with optional noisy layers for exploration.

    When `trunk` is provided, the MLP will be a wrapper around the trunk.

    See `Noisy Networks for Exploration (Fortunato et al., ICLR 2018)` for more details
    on the noisy layers.
    """

    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dim: int = 256,
        n_hidden_layers: Optional[int] = 2,
        n_noisy_layers: int = 0,
        activation_fn: Literal["relu", "leaky_relu", "tanh", "elu"] = "relu",
        trunk: Optional[nn.Module] = None,
        add_layer_norm: bool = False,
        std_init: float = 0.1,
    ):
        """Initializes a new MLP.

        Args:
            input_dim: The dimension of the input.
            output_dim: The dimension of the output.
            hidden_dim: The dimension of the hidden layers.
            n_hidden_layers: The number of hidden layers (including the input
                projection). With n_hidden_layers=2, the architecture is
                [input→H1→H2→output] (3 Linear layers total).
            n_noisy_layers: The number of noisy layers (from the output end).
                The input projection is never noisy. Must be <= n_hidden_layers.
            activation_fn: The activation function to use.
            trunk: A custom trunk to use. If None, a new trunk will be created.
            add_layer_norm: Whether to add layer normalization to the hidden layers.
            noise_std: The inital value of the noise standard deviation for noisy layers.
        """
        super().__init__()
        self._input_dim = input_dim
        self._output_dim = output_dim

        # Validate hidden layer inputs
        assert n_noisy_layers >= 0, "n_noisy_layers must be non-negative (>= 0)."
        assert std_init > 0, "std_init must be positive (> 0)."

        # If a trunk is provided, the MLP will be a wrapper around the trunk.
        if trunk is not None:
            assert (
                n_noisy_layers <= 1
            ), "when trunk is provided, n_noisy_layers must be 0 or 1."
            self.trunk = trunk
            assert hasattr(trunk, "hidden_dim") and isinstance(
                trunk.hidden_dim, torch.Tensor
            ), "trunk must have a hidden_dim attribute"
            self._hidden_dim = int(trunk.hidden_dim.item())
        # If no trunk is provided, we build the MLP from scratch. Noisy layers are added
        # to the end of the network.
        else:
            assert n_hidden_layers is not None, "n_hidden_layers must be provided"
            assert n_hidden_layers >= 1, "n_hidden_layers must be >= 1."
            # The input projection is always non-noisy. At most (n_hidden_layers - 1)
            # extra hidden layers plus the output layer can be noisy, giving a maximum
            # of n_hidden_layers noisy layers.
            assert (
                n_noisy_layers <= n_hidden_layers
            ), "n_noisy_layers must be <= n_hidden_layers."
            assert hidden_dim is not None, "hidden_dim must be provided"
            assert (
                activation_fn in ACTIVATION_FNS
            ), "activation_fn must be one of " + str(ACTIVATION_FNS.keys())

            activation = ACTIVATION_FNS[activation_fn]

            # Initialize the input layer (never noisy).
            if add_layer_norm:
                arch = [
                    nn.Linear(input_dim, hidden_dim),
                    nn.LayerNorm(hidden_dim),
                    activation(),
                ]
            else:
                arch = [
                    nn.Linear(input_dim, hidden_dim),
                    activation(),
                ]

            # Add hidden-to-hidden layers. The input projection above counts as
            # the first hidden layer, so we add (n_hidden_layers - 1) more.
            # Put all noisy layers near the output.
            n_extra_hidden = max(0, n_hidden_layers - 1)
            n_noisy_hidden_layers = max(0, n_noisy_layers - 1)
            n_noiseless_hidden_layers = n_extra_hidden - n_noisy_hidden_layers
            hidden_layer_types = [nn.Linear] * n_noiseless_hidden_layers + [
                NoisyLinear
            ] * n_noisy_hidden_layers

            for layer_type in hidden_layer_types:
                if isinstance(layer_type, NoisyLinear):
                    arch.append(layer_type(hidden_dim, hidden_dim, std_init=std_init))
                else:
                    arch.append(layer_type(hidden_dim, hidden_dim))

                if add_layer_norm:
                    arch.append(nn.LayerNorm(hidden_dim))

                arch.append(activation())

            self.trunk = nn.Sequential(*arch)
            self.trunk.hidden_dim = torch.tensor(hidden_dim)
            self._hidden_dim = hidden_dim

        # Initialize the output layer.
        if n_noisy_layers == 0:
            self.last_layer = nn.Linear(self._hidden_dim, output_dim)
        else:
            self.last_layer = NoisyLinear(
                self._hidden_dim, output_dim, std_init=std_init  # type: ignore
            )

    def forward(self, preprocessed_states: torch.Tensor) -> torch.Tensor:
        """Forward method for the neural network.

        Args:
            preprocessed_states: a batch of states appropriately preprocessed for
                ingestion by the MLP. The shape of the tensor should be (*batch_shape, input_dim).
        Returns: a tensor of shape (*batch_shape, output_dim).
        """
        if not preprocessed_states.is_floating_point():
            preprocessed_states = preprocessed_states.to(torch.get_default_dtype())

        out = self.trunk(preprocessed_states)
        out = self.last_layer(out)
        return out

    @property
    def input_dim(self):
        return self._input_dim

    @property
    def output_dim(self):
        return self._output_dim

    @input_dim.setter
    def input_dim(self, value: int):
        self._input_dim = value

    @output_dim.setter
    def output_dim(self, value: int):
        self._output_dim = value


class Tabular(nn.Module):
    """Implements a tabular policy.

    This class is only compatible with the EnumPreprocessor.

    Attributes:
        table: a tensor with dimensions [n_states, output_dim].
        device: the device that holds this policy.
    """

    def __init__(self, n_states: int, output_dim: int) -> None:
        """Initializes a new Tabular module.

        Args:
            n_states: The number of states.
            output_dim: The dimension of the output.
        """
        super().__init__()
        self.table = nn.parameter.Parameter(torch.zeros((n_states, output_dim)))
        self.device = None

    def forward(self, preprocessed_states: torch.Tensor) -> torch.Tensor:
        """Forward method for the tabular policy.

        Args:
            preprocessed_states: a batch of states appropriately preprocessed for
                ingestion by the tabular policy. The shape of the tensor should be (*batch_shape, 1).
        Returns: a tensor of shape (*batch_shape, output_dim).
        """
        if self.device is None:
            self.device = preprocessed_states.device
            self.table = self.table.to(self.device)
        assert is_int_dtype(preprocessed_states)
        outputs = self.table[preprocessed_states.squeeze(-1)]
        return outputs


class UniformModule(nn.Module):
    """Constant-output module for non-learned (uniform/fixed) policies.

    Outputs a constant tensor for all inputs.  Plug into any ``Estimator`` as the
    ``module`` argument to get a non-learned policy.

    Typical fill values:

    * ``0.0`` — equal logits → uniform categorical (discrete environments).
    * ``1.0`` — fixed params for sigmoid-based continuous estimators.

    Attributes:
        output_dim: Output dimension matching the estimator's expected_output_dim.
        fill_value: Constant value to fill the output tensor.
        skip_normalization: If True, signals to estimators that outputs should be
            used directly without normalization transforms (e.g. sigmoid scaling
            of concentration parameters).
    """

    def __init__(
        self,
        output_dim: int,
        input_dim: int | None = None,
        fill_value: float = 0.0,
        skip_normalization: bool = False,
    ) -> None:
        """Initialize a UniformModule.

        Args:
            output_dim: The dimension of the output.
            input_dim: Input dimension. Set this when no external preprocessor is
                provided so that ``Estimator`` can create a default
                ``IdentityPreprocessor``.  May be omitted when a preprocessor is
                always supplied.
            fill_value: Constant value for every output element.
            skip_normalization: If True, downstream estimators should skip
                normalization of the module output (e.g. don't apply sigmoid
                scaling to concentration parameters).
        """
        super().__init__()
        if input_dim is not None:
            self.input_dim = input_dim
        self.output_dim = output_dim
        self.fill_value = fill_value
        self.skip_normalization = skip_normalization

    def forward(self, preprocessed_states: torch.Tensor) -> torch.Tensor:
        """Return a constant tensor of shape ``(*batch_shape, output_dim)``.

        Args:
            preprocessed_states: Tensor of shape ``(*batch_shape, input_dim)``.

        Returns:
            Tensor of shape ``(*batch_shape, output_dim)`` filled with
            ``fill_value``.
        """
        return torch.full(
            (*preprocessed_states.shape[:-1], self.output_dim),
            self.fill_value,
            device=preprocessed_states.device,
        )


class DiscreteUniform(UniformModule):
    """Uniform distribution over discrete actions.

    Backward-compatible alias for ``UniformModule(output_dim, fill_value=0.0)``.
    Prefer ``UniformModule`` for new code.
    """

    def __init__(self, output_dim: int) -> None:
        super().__init__(output_dim=output_dim, fill_value=0.0)


class LinearTransformer(nn.Module):
    """The Linear Transformer module.

    Implements Transformers are RNNs: Fast Autoregressive Transformers with Linear
        Attention. Angelos Katharopoulos, Apoorv Vyas, Nikolaos Pappas, François
        Fleuret, ICML 2020.

    Expresses self-attention as a linear dot-product of kernel feature maps and makes
    use the associativity property of matrix products to reduce the complexity of the
    attention computation from O(n^2) to O(n).

    Implementation from https://github.com/lucidrains/linear-attention-transformer.

    Args:
        dim: The dimension of the input.
        depth: The depth of the transformer.
        max_seq_len: The maximum sequence length.
        n_heads: The number of attention heads.
        causal: Whether to use causal attention.
    """

    def __init__(
        self,
        dim: int,
        depth: int,
        max_seq_len: int,
        n_heads: int = 8,
        causal: bool = False,
    ):
        """Initializes a new LinearTransformer module.

        Args:
            dim: The dimension of the input.
            depth: The depth of the transformer.
            max_seq_len: The maximum sequence length.
            n_heads: The number of attention heads.
            causal: Whether to use causal attention.
        """
        super().__init__()
        assert isinstance(dim, int) and dim > 0, "dim must be a positive integer"
        assert isinstance(depth, int) and depth > 0, "depth must be a positive integer"
        assert (
            isinstance(max_seq_len, int) and max_seq_len > 0
        ), "max_seq_len must be a positive integer"

        self.module = LinearAttentionTransformer(
            dim,
            depth,
            max_seq_len,
            heads=n_heads,
            causal=causal,
            dim_head=None,
            bucket_size=64,
            ff_chunks=1,
            ff_glu=False,
            ff_dropout=0.0,
            attn_layer_dropout=0.0,
            attn_dropout=0.0,
            reversible=False,
            blindspot_size=1,
            n_local_attn_heads=0,
            local_attn_window_size=128,
            receives_context=False,
            attend_axially=False,
            pkm_layers=tuple(),
            pkm_num_keys=128,
            linformer_settings=None,
            context_linformer_settings=None,
            shift_tokens=False,
        )
        # TODO: Should we have a final linear layer as part of this module?
        # The output dimension is the same as the embedding dimension.
        self.output_dim = dim

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.module(x)


class GraphActionGNN(nn.Module):
    """Implements a GNN for graph action prediction."""

    def __init__(
        self,
        num_node_classes: int,
        num_edge_classes: int,
        directed: bool,
        embedding_dim: int = 128,
        num_conv_layers: int = 1,
        is_backward: bool = False,
    ) -> None:
        """Initializes a new GraphActionGNN module.

        Args:
            n_nodes: The number of nodes in the graph.
            directed: Whether the graph is directed.
            num_edge_classes: The number of edge classes.
            num_conv_layers: The number of convolutional layers.
            embedding_dim: The dimension of the node embeddings.
            is_backward: Whether the GNN is used for a backward policy.
        """
        super().__init__()
        assert num_node_classes > 0, "num_node_classes must be greater than 0"
        assert embedding_dim > 0, "embedding_dim must be greater than 0"
        assert num_conv_layers > 0, "num_conv_layers must be greater than 0"
        assert isinstance(num_node_classes, int), "n_nodes must be an integer"
        assert isinstance(embedding_dim, int), "embedding_dim must be an integer"
        assert isinstance(num_conv_layers, int), "num_conv_layers must be an integer"
        assert isinstance(directed, bool), "directed must be a boolean"
        assert isinstance(is_backward, bool), "is_backward must be a boolean"
        self.num_node_classes = num_node_classes
        self.hidden_dim = self.embedding_dim = embedding_dim
        self.is_backward = is_backward
        self.is_directed = directed
        self.num_conv_layers = num_conv_layers
        self.num_edge_classes = num_edge_classes

        # Node embedding layer.
        self.embedding = nn.Embedding(num_node_classes, self.embedding_dim)
        self.conv_blks = nn.ModuleList()

        if directed:
            for i in range(num_conv_layers):
                self.conv_blks.extend(
                    [
                        DirGNNConv(
                            GCNConv(
                                self.embedding_dim if i == 0 else self.hidden_dim,
                                self.hidden_dim,
                            ),
                            alpha=0.5,
                            root_weight=True,
                        ),
                        # Process in/out components separately
                        nn.ModuleList(
                            [
                                nn.Sequential(
                                    nn.Linear(
                                        self.hidden_dim // 2,
                                        self.hidden_dim // 2,
                                    ),
                                    nn.ReLU(),
                                    nn.Linear(
                                        self.hidden_dim // 2,
                                        self.hidden_dim // 2,
                                    ),
                                )
                                for _ in range(2)  # 1 for in & 1 for out-features.
                            ]
                        ),
                    ]
                )
        else:  # Undirected case.
            for i in range(num_conv_layers):
                self.conv_blks.extend(
                    [
                        GINConv(
                            MLP(
                                input_dim=(
                                    self.embedding_dim if i == 0 else self.hidden_dim
                                ),
                                output_dim=self.hidden_dim,
                                hidden_dim=self.hidden_dim,
                                n_hidden_layers=1,
                                add_layer_norm=True,
                            ),
                        ),
                        nn.Sequential(
                            nn.Linear(self.hidden_dim, self.hidden_dim),
                            nn.ReLU(),
                            nn.Linear(self.hidden_dim, self.hidden_dim),
                        ),
                    ]
                )

        self.norm = nn.LayerNorm(self.hidden_dim)

        # Heads operating on per-graph pooled features
        self.action_type_mlp = MLP(
            input_dim=self.hidden_dim,
            output_dim=3,
            hidden_dim=self.hidden_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )
        self.node_class_mlp = MLP(
            input_dim=self.hidden_dim,
            output_dim=self.num_node_classes,
            hidden_dim=self.hidden_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )
        self.node_index_mlp = MLP(
            input_dim=self.hidden_dim,
            output_dim=1,
            hidden_dim=self.hidden_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )
        self.edge_class_mlp = MLP(
            input_dim=self.hidden_dim,
            output_dim=self.num_edge_classes,
            hidden_dim=self.hidden_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )

    @property
    def input_dim(self):
        return 1  # placeholder TODO: remove this

    @staticmethod
    def _group_mean(tensor: torch.Tensor, batch_ptr: torch.Tensor) -> torch.Tensor:
        # Safe mean over ragged graphs using ptr; returns zeros for empty graphs
        if tensor.numel() == 0:
            B = batch_ptr.numel() - 1
            return torch.zeros(B, 0, device=batch_ptr.device)
        cumsum = torch.zeros(
            (len(tensor) + 1, tensor.size(-1)), dtype=tensor.dtype, device=tensor.device
        )
        cumsum[1:] = torch.cumsum(tensor, dim=0)
        size = batch_ptr[1:] - batch_ptr[:-1]
        denom = torch.clamp(size, min=1).to(tensor.dtype)
        sums = cumsum[batch_ptr[1:]] - cumsum[batch_ptr[:-1]]
        means = sums / denom[:, None]
        means[size == 0] = 0
        return means

    def forward(self, states_tensor: GeometricBatch) -> TensorDict:
        node_features = states_tensor.x
        B = len(states_tensor)
        lengths = states_tensor.ptr[1:] - states_tensor.ptr[:-1]
        max_nodes = int(lengths.max().item())
        device = node_features.device

        # Embed node classes
        if node_features.numel() > 0:
            x = self.embedding(node_features.squeeze(-1))
        # Handle the case where the graph has no nodes. We use zeros as
        # features, so we can continue the forward pass.
        else:
            x = torch.zeros(0, self.hidden_dim, device=device)

        # Message passing with residual connections and layer norms
        for i in range(0, len(self.conv_blks), 2):
            x_new = self.conv_blks[i](x, states_tensor.edge_index)  # GIN/GCN conv.
            if self.is_directed:
                assert isinstance(self.conv_blks[i + 1], nn.ModuleList)
                x_in, x_out = torch.chunk(x_new, 2, dim=-1)

                # Process each component separately through its own MLP.
                mlp_in, mlp_out = self.conv_blks[i + 1]
                x_in = mlp_in(x_in)
                x_out = mlp_out(x_out)
                x_new = torch.cat([x_in, x_out], dim=-1)
            else:
                x_new = self.conv_blks[i + 1](x_new)  # Linear -> ReLU -> Linear.

            x = x_new + x if i > 0 else x_new  # Residual connection.
            x = self.norm(x)  # Layernorm.

            # Pool to graph features
        graph_emb = (
            self._group_mean(x, states_tensor.ptr)
            if x.numel() > 0
            else torch.zeros(B, self.hidden_dim, device=device)
        )

        # Action type and class logits
        action_type = self.action_type_mlp(graph_emb)
        node_class_logits = self.node_class_mlp(graph_emb)
        edge_class_logits = self.edge_class_mlp(graph_emb)

        # Node index logits
        if self.is_backward:
            node_index_logits = self.node_index_mlp(x).squeeze(-1)
            node_index_logits = torch.split(node_index_logits, lengths.tolist(), dim=0)
            node_index_logits = torch.nn.utils.rnn.pad_sequence(
                list(node_index_logits), batch_first=True
            )
        else:
            node_index_logits = torch.zeros(B, max_nodes + 1, device=device)

        # Edge-index logits via pairwise dot products
        # Pad to max_nodes across batch for gathering candidate edges
        seqs = torch.split(x, lengths.tolist())
        padded = torch.nn.utils.rnn.pad_sequence(
            list(seqs), batch_first=True
        )  # (B, max_nodes, hidden_dim)

        feature_dim = (self.hidden_dim // 2) if self.is_directed else self.hidden_dim
        if self.is_directed:
            source_features = padded[..., :feature_dim]
            target_features = padded[..., feature_dim:]
            scores = torch.einsum("bnf,bmf->bnm", source_features, target_features)
        else:
            scores = torch.einsum("bnf,bmf->bnm", padded, padded)
        scores = scores / math.sqrt(max(1, feature_dim))

        ei0, ei1 = get_edge_indices(max_nodes, self.is_directed, device)
        edge_index_logits = scores[:, ei0, ei1]

        return TensorDict(
            {
                GraphActions.ACTION_TYPE_KEY: action_type,
                GraphActions.NODE_CLASS_KEY: node_class_logits,
                GraphActions.NODE_INDEX_KEY: node_index_logits,
                GraphActions.EDGE_CLASS_KEY: edge_class_logits,
                GraphActions.EDGE_INDEX_KEY: edge_index_logits,
            },
            batch_size=B,
        )


class GraphEdgeActionMLP(nn.Module):
    """Network that processes flattened adjacency matrices to predict graph actions.

    Unlike the GNN-based GraphActionGNN, this module uses standard MLPs to process
    the entire adjacency matrix as a flattened vector. This approach:

    1. Can directly process global graph structure without message passing.
    2. May be more effective for small graphs where global patterns are important.
    3. Does not require complex graph neural network operations.

    The module architecture consists of:
    - An MLP to process the flattened adjacency matrix into an embedding.
    - An edge MLP that predicts logits for each possible edge action.
    - An exit MLP that predicts a logit for the exit action.

    Args:
        n_nodes: Number of nodes in the graph.
        directed: Whether the graph is directed or undirected.
        n_hidden_layers: Number of hidden layers in the MLP for the edge actions.
        n_hidden_layers_exit: Number of hidden layers in the MLP for the exit action.
        embedding_dim: Dimension of internal embeddings.
        is_backward: Whether this is a backward policy.
    """

    def __init__(
        self,
        n_nodes: int,
        directed: bool,
        num_node_classes: int,
        num_edge_classes: int,
        n_hidden_layers: int = 2,
        n_hidden_layers_exit: int = 1,
        embedding_dim: int = 128,
        is_backward: bool = False,
    ):
        """Initializes a new GraphEdgeActionMLP module.

        Args:
            n_nodes: The number of nodes in the graph.
            directed: Whether the graph is directed.
            num_node_classes: The number of node classes.
            num_edge_classes: The number of edge classes.
            n_hidden_layers: The number of hidden layers in the main MLP.
            n_hidden_layers_exit: The number of hidden layers in the exit MLP.
            embedding_dim: The dimension of the embeddings.
            is_backward: Whether the MLP is used for a backward policy.
        """
        super().__init__()
        assert n_nodes > 0, "n_nodes must be greater than 0"
        assert embedding_dim > 0, "embedding_dim must be greater than 0"
        assert n_hidden_layers > 0, "n_hidden_layers must be greater than 0"
        assert n_hidden_layers_exit > 0, "n_hidden_layers_exit must be greater than 0"
        assert isinstance(n_nodes, int), "n_nodes must be an integer"
        assert isinstance(embedding_dim, int), "embedding_dim must be an integer"
        assert isinstance(n_hidden_layers, int), "n_hidden_layers must be an integer"
        assert isinstance(
            n_hidden_layers_exit, int
        ), "n_hidden_layers_exit must be an integer"
        assert isinstance(directed, bool), "directed must be a boolean"
        assert isinstance(is_backward, bool), "is_backward must be a boolean"
        self._input_dim = n_nodes**2
        self.n_nodes = n_nodes
        self.is_directed = directed
        self.is_backward = is_backward
        self.hidden_dim = embedding_dim
        self.num_node_classes = num_node_classes
        self.num_edge_classes = num_edge_classes

        # MLP for processing the flattened adjacency matrix
        self.mlp = MLP(
            input_dim=n_nodes**2,  # Flattened adjacency matrix
            output_dim=embedding_dim,
            hidden_dim=embedding_dim,
            n_hidden_layers=n_hidden_layers,
            add_layer_norm=True,
        )

        # Exit action MLP
        self.exit_mlp = MLP(
            input_dim=embedding_dim,
            output_dim=1,
            hidden_dim=embedding_dim,
            n_hidden_layers=n_hidden_layers_exit,
            add_layer_norm=True,
        )

        # Edge prediction MLP
        # Output dimension.
        edges_dim = self.n_nodes**2 - self.n_nodes
        if not self.is_directed:
            edges_dim = edges_dim // 2  # No double-counting.

        if not self.is_backward:
            out_dim = edges_dim + 1  # +1 for exit action.
        else:
            out_dim = edges_dim

        self._output_dim = int(out_dim)
        self._edges_dim = int(edges_dim)

        self.node_class_mlp = MLP(
            input_dim=embedding_dim,
            output_dim=self.num_node_classes,
            hidden_dim=embedding_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )
        self.features_embedding = nn.Embedding(self.num_node_classes, embedding_dim)
        self.node_index_mlp = MLP(
            input_dim=embedding_dim,
            output_dim=1,
            hidden_dim=embedding_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )
        self.edge_mlp = MLP(
            input_dim=embedding_dim,
            output_dim=self.edges_dim,
            hidden_dim=embedding_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )

        self.edge_class_mlp = MLP(
            input_dim=embedding_dim,
            output_dim=self.num_edge_classes,
            hidden_dim=embedding_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )

    @property
    def input_dim(self) -> int:
        return self._input_dim

    @property
    def output_dim(self) -> int:
        return self._output_dim

    @property
    def edges_dim(self) -> int:
        return self._edges_dim

    def forward(self, states_tensor: GeometricBatch) -> TensorDict:
        """Forward pass to compute action logits from graph states.

        Process:
        1. Convert the graph representation to adjacency matrices
        2. Process the flattened adjacency matrices through the main MLP
        3. Predict logits for edge actions and exit action

        Args:
            states_tensor: A GeometricBatch containing graph state information

        Returns:
            A tensor of logits for all possible actions
        """
        device = states_tensor.x.device
        # Convert the graph to adjacency matrix.
        adj_matrices = torch.zeros(
            (len(states_tensor), self.n_nodes, self.n_nodes),
            device=device,
        )

        # Fill the adjacency matrices from edge indices
        if states_tensor.edge_index.numel() > 0:
            for i in range(len(states_tensor)):
                eis = states_tensor[i].edge_index
                adj_matrices[i, eis[0], eis[1]] = 1

        # Flatten the adjacency matrices for the MLP
        adj_matrices_flat = adj_matrices.view(len(states_tensor), -1)

        # Process with MLP
        embedding = self.mlp(adj_matrices_flat)

        # Generate edge and exit actions
        edge_actions = self.edge_mlp(embedding)

        action_type = torch.ones(len(states_tensor), 3, device=device) * float("-inf")
        edge_class_logits = torch.zeros(
            len(states_tensor), self.num_edge_classes, device=device
        )
        lengths = states_tensor.ptr[1:] - states_tensor.ptr[:-1]
        if self.is_backward:
            action_type[..., GraphActionType.ADD_EDGE] = 0.0
            node_features = self.features_embedding(states_tensor.x)
            node_index_logits = self.node_index_mlp(node_features).squeeze(-1)
            node_index_logits = torch.split(node_index_logits, lengths.tolist(), dim=0)
            node_index_logits = torch.nn.utils.rnn.pad_sequence(
                list(node_index_logits), batch_first=True
            )
        else:
            exit_action = self.exit_mlp(embedding).squeeze(-1)
            action_type[..., GraphActionType.ADD_EDGE] = 0.0
            action_type[..., GraphActionType.EXIT] = exit_action
            edge_class_logits = self.edge_class_mlp(embedding)
            node_class_logits = self.node_class_mlp(embedding)
            node_index_logits = torch.zeros(
                len(states_tensor), self.n_nodes + 1, device=device
            )

        return TensorDict(
            {
                GraphActions.ACTION_TYPE_KEY: action_type,
                GraphActions.NODE_CLASS_KEY: node_class_logits,
                GraphActions.NODE_INDEX_KEY: node_index_logits,
                GraphActions.EDGE_CLASS_KEY: edge_class_logits,
                GraphActions.EDGE_INDEX_KEY: edge_actions,
            },
            batch_size=len(states_tensor),
        )


class GraphScalarMLP(nn.Module):
    """Graph encoder that maps adjacency structure to n scalar output."""

    def __init__(
        self,
        n_nodes: int,
        directed: bool,
        embedding_dim: int = 128,
        n_hidden_layers: int = 2,
        n_outputs: int = 1,
    ) -> None:
        super().__init__()
        assert n_nodes > 0, "n_nodes must be positive"
        assert embedding_dim > 0, "embedding_dim must be positive"
        assert n_hidden_layers >= 1, "n_hidden_layers must be >= 1"
        self.n_nodes = n_nodes
        self.is_directed = directed
        self.input_dim = n_nodes**2

        self.backbone = MLP(
            input_dim=n_nodes**2,
            output_dim=embedding_dim,
            hidden_dim=embedding_dim,
            n_hidden_layers=n_hidden_layers,
            add_layer_norm=True,
        )
        self.head = MLP(
            input_dim=embedding_dim,
            output_dim=n_outputs,
            hidden_dim=embedding_dim,
            n_hidden_layers=1,
            add_layer_norm=True,
        )

    def forward(self, states_tensor: GeometricBatch) -> torch.Tensor:
        """Encode graphs into a scalar per element of the batch."""
        batch_size = len(states_tensor)
        device = states_tensor.x.device
        adj = torch.zeros((batch_size, self.n_nodes, self.n_nodes), device=device)

        if states_tensor.edge_index.numel() > 0:
            # Vectorized over the batch dim.
            B = torch.arange(batch_size)
            edges = states_tensor.edge_index
            adj[B, edges[:, 0], edges[:, 1]] = 1
            if not self.is_directed:
                adj[B, edges[:, 1], edges[:, 0]] = 1

        embedding = self.backbone(adj.view(batch_size, -1))
        return self.head(embedding)


class GraphActionUniform(nn.Module):
    """Implements a uniform distribution over discrete actions given a graph state.

    It uses a zero function approximator (a function that always outputs 0) to be used as
    logits by a DiscretePBEstimator.

    Attributes:
        output_dim: The size of the output space.
    """

    def __init__(
        self,
        edges_dim: int,
        num_edge_classes: int,
        num_node_classes: int,
    ) -> None:
        """Initializes a new GraphActionUniform module.

        Args:
            edges_dim: The dimension of edge_index in GraphActions.
            num_edge_classes: The number of edge classes.
            num_node_classes: The number of node classes.
        """
        """Initializes the uniform function approximiator.

        Args:
            edges_dim (int): The dimension of edge_index in GraphActions.
            num_edge_classes (int): Number of edge classes.
            num_node_classes (int): Number of node classes.
        """
        super().__init__()
        self.input_dim = 1  # has no effect
        self.edges_dim = edges_dim
        self.num_edge_classes = num_edge_classes
        self.num_node_classes = num_node_classes

    def forward(self, states_tensor: GeometricBatch) -> TensorDict:
        """Forward method for the uniform distribution.

        Args:
            states_tensor: a batch of states appropriately preprocessed for
                ingestion by the uniform distribution.

        Returns:
            A TensorDict containing logits for each action component, with all values
              set to 1 to represent a uniform distribution:
            - GraphActions.ACTION_TYPE_KEY: Tensor of shape [*batch_shape, 3] for the 3
              possible action types
            - GraphActions.EDGE_CLASS_KEY: Tensor of shape [*batch_shape,
              num_edge_classes] for edge class logits
            - GraphActions.NODE_CLASS_KEY: Tensor of shape [*batch_shape,
              num_node_classes] for node class logits
            - GraphActions.EDGE_INDEX_KEY: Tensor of shape [*batch_shape, edges_dim]
              for edge index logits
        """
        device = states_tensor.x.device
        max_nodes = int(torch.max(states_tensor.ptr[1:] - states_tensor.ptr[:-1]))
        return TensorDict(
            {
                GraphActions.ACTION_TYPE_KEY: torch.ones(
                    len(states_tensor), 3, device=device
                ),
                GraphActions.EDGE_CLASS_KEY: torch.ones(
                    len(states_tensor), self.num_edge_classes, device=device
                ),
                GraphActions.NODE_CLASS_KEY: torch.ones(
                    len(states_tensor), self.num_node_classes, device=device
                ),
                GraphActions.NODE_INDEX_KEY: torch.ones(
                    len(states_tensor), max_nodes, device=device
                ),
                GraphActions.EDGE_INDEX_KEY: torch.ones(
                    len(states_tensor), self.edges_dim, device=device
                ),
            },
            batch_size=len(states_tensor),
        )


class NoisyLinear(nn.Linear):
    """Noisy Linear Layer.

    Presented in "Noisy Networks for Exploration", https://arxiv.org/abs/1706.10295v3

    A Noisy Linear Layer is a linear layer with parametric noise added to the weights.
    This induced stochasticity can be used in RL networks for the agent's policy to aid
    efficient exploration. The parameters of the noise are learned with gradient descent
    along with any other remaining network weights. Factorized Gaussian noise is the
    type of noise usually employed.

    Taken from torchrl v0.9.2.

    Args:
        in_features (int): input features dimension
        out_features (int): out features dimension
        bias (bool, optional): if ``True``, a bias term will be added to the matrix multiplication: Ax + b.
            Defaults to ``True``
        device (DEVICE_TYPING, optional): device of the layer.
            Defaults to ``"cpu"``
        dtype (torch.dtype, optional): dtype of the parameters.
            Defaults to ``None`` (default pytorch dtype)
        std_init (scalar, optional): initial value of the Gaussian standard deviation before optimization.
            Defaults to ``0.1``

    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = True,
        device: torch.device | None = None,
        dtype: torch.dtype | None = None,
        std_init: float = 0.1,
    ):
        nn.Module.__init__(self)
        self.in_features = int(in_features)
        self.out_features = int(out_features)
        self.std_init = std_init

        self.weight_mu = nn.Parameter(
            torch.empty(
                out_features,
                in_features,
                device=device,
                dtype=dtype,
                requires_grad=True,
            )
        )
        self.weight_sigma = nn.Parameter(
            torch.empty(
                out_features,
                in_features,
                device=device,
                dtype=dtype,
                requires_grad=True,
            )
        )
        self.register_buffer(
            "weight_epsilon",
            torch.empty(out_features, in_features, device=device, dtype=dtype),
        )
        if bias:
            self.bias_mu = nn.Parameter(
                torch.empty(
                    out_features,
                    device=device,
                    dtype=dtype,
                    requires_grad=True,
                )
            )
            self.bias_sigma = nn.Parameter(
                torch.empty(
                    out_features,
                    device=device,
                    dtype=dtype,
                    requires_grad=True,
                )
            )
            self.register_buffer(
                "bias_epsilon",
                torch.empty(out_features, device=device, dtype=dtype),
            )
        else:
            self.bias_mu = None
        self.reset_parameters()
        self.reset_noise()

    def reset_parameters(self) -> None:
        mu_range = 1 / math.sqrt(self.in_features)
        self.weight_mu.data.uniform_(-mu_range, mu_range)
        self.weight_sigma.data.fill_(self.std_init / math.sqrt(self.in_features))
        if self.bias_mu is not None:
            self.bias_mu.data.uniform_(-mu_range, mu_range)
            self.bias_sigma.data.fill_(self.std_init / math.sqrt(self.out_features))

    def reset_noise(self) -> None:
        epsilon_in = self._scale_noise(self.in_features)
        epsilon_out = self._scale_noise(self.out_features)
        self.weight_epsilon.copy_(epsilon_out.outer(epsilon_in))  # type: ignore
        if self.bias_mu is not None:
            self.bias_epsilon.copy_(epsilon_out)  # type: ignore

    def _scale_noise(self, size: int | torch.Size) -> torch.Tensor:
        if isinstance(size, int):
            size = (size,)  # type: ignore
        x = torch.randn(*size, device=self.weight_mu.device)
        return x.sign().mul_(x.abs().sqrt_())

    @property
    def weight(self) -> torch.Tensor:
        if self.training:
            return self.weight_mu + self.weight_sigma * self.weight_epsilon  # type: ignore
        else:
            return self.weight_mu

    @property
    def bias(self) -> torch.Tensor | None:
        if self.bias_mu is not None:
            if self.training:
                return self.bias_mu + self.bias_sigma * self.bias_epsilon  # type: ignore
            else:
                return self.bias_mu
        else:
            return None


class AutoregressiveDiscreteSequenceModel(ABC, nn.Module):

    @abstractmethod
    def init_carry(
        self,
        batch_size: int,
        device: torch.device,
    ) -> dict[str, torch.Tensor]:
        """Initialize the carry for the sequence model.

        Args:
            batch_size (int): Batch size.
            device (torch.device): Device to allocate carry tensors on.

        Returns:
            dict[str, torch.Tensor]: Initialized carry.
        """

    @abstractmethod
    def forward(
        self,
        x: torch.Tensor,
        carry: dict[str, torch.Tensor],
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        """Compute the logits for the next tokens in the sequence.

        Args:
            x (torch.Tensor): (B, T) tensor of input token indices where ``T`` is the
                number of newly supplied timesteps (``T`` may be 1 for incremental
                decoding).
            carry (dict[str, torch.Tensor]): Carry from previous steps for recurrent
                processing (e.g., hidden states).

        Returns:
            tuple[torch.Tensor, dict[str, torch.Tensor]]: Logits for the next token
                at each supplied timestep with shape (B, T, vocab) and updated carry.
        """

    @property
    @abstractmethod
    def vocab_size(self) -> int:
        """Size of the vocabulary (excluding BOS token)."""


class RecurrentDiscreteSequenceModel(AutoregressiveDiscreteSequenceModel):
    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int,
        hidden_size: int,
        num_layers: int = 1,
        rnn_type: Literal["lstm", "gru"] = "lstm",
        dropout: float = 0.0,
    ) -> None:
        super().__init__()
        if num_layers <= 0:
            raise ValueError("num_layers must be a positive integer.")
        rnn_kind = rnn_type.lower()
        if rnn_kind not in {"lstm", "gru"}:
            raise ValueError("rnn_type must be 'lstm' or 'gru'.")

        if not 0.0 <= dropout <= 1.0:
            raise ValueError("dropout must be in the range [0, 1].")

        self._vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        self.rnn_type = rnn_kind

        self.embedding = nn.Embedding(vocab_size + 1, embedding_dim)  # +1 for BOS token
        rnn_dropout = dropout if num_layers > 1 else 0.0
        self.lstm: nn.LSTM | None
        self.gru: nn.GRU | None
        if rnn_kind == "lstm":
            self.lstm = nn.LSTM(
                input_size=embedding_dim,
                hidden_size=hidden_size,
                num_layers=num_layers,
                batch_first=True,
                dropout=rnn_dropout,
            )
            self.gru = None
        else:
            self.gru = nn.GRU(
                input_size=embedding_dim,
                hidden_size=hidden_size,
                num_layers=num_layers,
                batch_first=True,
                dropout=rnn_dropout,
            )
            self.lstm = None
        self.output_projection = nn.Linear(hidden_size, vocab_size)

    def init_carry(
        self,
        batch_size: int,
        device: torch.device,
    ) -> dict[str, torch.Tensor]:
        carry: dict[str, torch.Tensor] = {
            "hidden": torch.zeros(
                self.num_layers, batch_size, self.hidden_size, device=device
            ),
        }
        if self.rnn_type == "lstm":
            carry["cell"] = torch.zeros(
                self.num_layers, batch_size, self.hidden_size, device=device
            )
        return carry

    def forward(
        self,
        x: torch.Tensor,
        carry: dict[str, torch.Tensor],
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        if x.dim() != 2:
            raise ValueError("Expected input tensor with shape (batch, timesteps).")

        batch, _timesteps = x.size()
        device = x.device

        if "hidden" not in carry:
            raise KeyError("Carry must provide a 'hidden' state tensor.")

        hidden = carry["hidden"]
        if hidden.size(1) != batch:
            raise ValueError(
                "Hidden state batch dimension does not match the provided tokens."
            )
        if hidden.device != device:
            raise ValueError(
                "Hidden state tensor must live on the same device as input tokens."
            )

        embedded = self.embedding(x)

        if self.rnn_type == "lstm":
            lstm = self.lstm
            if lstm is None:
                raise RuntimeError("LSTM module was not initialized.")
            if "cell" not in carry:
                raise KeyError("LSTM carry must provide a 'cell' state tensor.")
            cell = carry["cell"]
            if cell.size(1) != batch:
                raise ValueError(
                    "Cell state batch dimension does not match the provided tokens."
                )
            if cell.device != device:
                raise ValueError(
                    "Cell state tensor must live on the same device as input tokens."
                )
            outputs, (hidden_next, cell_next) = lstm(embedded, (hidden, cell))
            updated_carry: dict[str, torch.Tensor] = {
                "hidden": hidden_next,
                "cell": cell_next,
            }
        else:
            gru = self.gru
            if gru is None:
                raise RuntimeError("GRU module was not initialized.")
            outputs, hidden_next = gru(embedded, hidden)
            updated_carry = {
                "hidden": hidden_next,
            }

        logits = self.output_projection(outputs)
        return logits, updated_carry

    @property
    def vocab_size(self) -> int:
        return self._vocab_size


class _AutoregressiveTransformerBlock(nn.Module):
    def __init__(
        self,
        embed_dim: int,
        num_heads: int,
        ff_hidden_dim: int,
        dropout: float,
    ) -> None:
        super().__init__()
        if embed_dim % num_heads != 0:
            raise ValueError("Embedding dimension must be divisible by number of heads.")

        self.embed_dim = embed_dim
        self.num_heads = num_heads
        self.head_dim = embed_dim // num_heads

        self.norm1 = nn.LayerNorm(embed_dim)
        self.norm2 = nn.LayerNorm(embed_dim)

        self.q_proj = nn.Linear(embed_dim, embed_dim)
        self.k_proj = nn.Linear(embed_dim, embed_dim)
        self.v_proj = nn.Linear(embed_dim, embed_dim)
        self.out_proj = nn.Linear(embed_dim, embed_dim)

        self.linear1 = nn.Linear(embed_dim, ff_hidden_dim)
        self.linear2 = nn.Linear(ff_hidden_dim, embed_dim)

        self.attn_dropout = nn.Dropout(dropout)
        self.residual_dropout = nn.Dropout(dropout)
        self.ff_dropout = nn.Dropout(dropout)

    def forward(
        self,
        hidden: torch.Tensor,
        key_carry: torch.Tensor,
        value_carry: torch.Tensor,
    ) -> tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        batch, timesteps, _ = hidden.size()

        normed_hidden = self.norm1(hidden)

        q = self.q_proj(normed_hidden)
        k = self.k_proj(normed_hidden)
        v = self.v_proj(normed_hidden)

        q = q.view(batch, timesteps, self.num_heads, self.head_dim).transpose(1, 2)
        k = k.view(batch, timesteps, self.num_heads, self.head_dim).transpose(1, 2)
        v = v.view(batch, timesteps, self.num_heads, self.head_dim).transpose(1, 2)

        carry_length = key_carry.size(2)
        updated_key_carry = torch.cat((key_carry, k), dim=2)
        updated_value_carry = torch.cat((value_carry, v), dim=2)

        attn_scores = torch.matmul(q, updated_key_carry.transpose(-2, -1)) / math.sqrt(
            float(self.head_dim)
        )

        if timesteps > 1 or carry_length > 0:
            total_kv_length = carry_length + timesteps
            kv_positions = torch.arange(
                total_kv_length, device=hidden.device, dtype=torch.long
            )
            query_positions = torch.arange(
                timesteps, device=hidden.device, dtype=torch.long
            ).unsqueeze(1)
            causal_mask = kv_positions.unsqueeze(0) <= (query_positions + carry_length)
            attn_scores = attn_scores.masked_fill(
                ~causal_mask.unsqueeze(0).unsqueeze(0), float("-inf")
            )

        attn_weights = torch.softmax(attn_scores, dim=-1)
        attn_weights = self.attn_dropout(attn_weights)
        attn_output = torch.matmul(attn_weights, updated_value_carry)
        attn_output = attn_output.transpose(1, 2).reshape(
            batch, timesteps, self.embed_dim
        )
        attn_output = self.out_proj(attn_output)

        residual = hidden
        hidden = residual + self.residual_dropout(attn_output)

        ff_input = self.norm2(hidden)
        ff_hidden = self.linear1(ff_input)
        ff_hidden = self.ff_dropout(F.gelu(ff_hidden))
        ff_hidden = self.linear2(ff_hidden)

        hidden = hidden + self.residual_dropout(ff_hidden)
        return hidden, updated_key_carry, updated_value_carry


class TransformerDiscreteSequenceModel(AutoregressiveDiscreteSequenceModel):
    def __init__(
        self,
        vocab_size: int,
        embedding_dim: int,
        num_heads: int,
        ff_hidden_dim: int,
        num_layers: int,
        max_position_embeddings: int,
        dropout: float = 0.0,
        positional_embedding: Literal["learned", "sinusoidal"] = "learned",
    ) -> None:
        super().__init__()
        if num_layers <= 0:
            raise ValueError("num_layers must be positive.")
        if max_position_embeddings <= 0:
            raise ValueError("max_position_embeddings must be positive.")
        if not 0.0 <= dropout <= 1.0:
            raise ValueError("dropout must lie in [0, 1].")
        if embedding_dim % num_heads != 0:
            raise ValueError("embedding_dim must be divisible by num_heads.")
        if positional_embedding not in {"learned", "sinusoidal"}:
            raise ValueError("positional_embedding must be 'learned' or 'sinusoidal'.")

        self._vocab_size = vocab_size
        self.embedding_dim = embedding_dim
        self.num_heads = num_heads
        self.ff_hidden_dim = ff_hidden_dim
        self.num_layers = num_layers
        self.max_position_embeddings = max_position_embeddings
        self.head_dim = embedding_dim // num_heads
        self._positional_embedding_type = positional_embedding

        self.token_embedding = nn.Embedding(
            vocab_size + 1, embedding_dim
        )  # +1 for BOS token
        if self._positional_embedding_type == "learned":
            self.position_embedding = nn.Embedding(
                max_position_embeddings, embedding_dim
            )
        else:
            self.position_embedding = SinusoidalPositionalEmbedding(
                embedding_dim=embedding_dim,
                max_length=max_position_embeddings,
            )
        self.embedding_dropout = nn.Dropout(dropout)

        blocks: list[_AutoregressiveTransformerBlock] = []
        for _ in range(num_layers):
            blocks.append(
                _AutoregressiveTransformerBlock(
                    embed_dim=embedding_dim,
                    num_heads=num_heads,
                    ff_hidden_dim=ff_hidden_dim,
                    dropout=dropout,
                )
            )

        self.layers = nn.ModuleList(blocks)
        self.final_norm = nn.LayerNorm(embedding_dim)
        self.output_projection = nn.Linear(embedding_dim, vocab_size)
        self.key_names = [f"key_{idx}" for idx in range(num_layers)]
        self.value_names = [f"value_{idx}" for idx in range(num_layers)]

    def init_carry(
        self,
        batch_size: int,
        device: torch.device,
    ) -> dict[str, torch.Tensor]:
        weight = self.token_embedding.weight
        carry: dict[str, torch.Tensor] = {
            "position": torch.zeros(batch_size, dtype=torch.long, device=device),
        }
        empty_key = weight.new_empty(batch_size, self.num_heads, 0, self.head_dim).to(
            device
        )
        empty_value = weight.new_empty(batch_size, self.num_heads, 0, self.head_dim).to(
            device
        )
        for key_name, value_name in zip(self.key_names, self.value_names):
            carry[key_name] = empty_key.clone()
            carry[value_name] = empty_value.clone()

        return carry

    def forward(
        self,
        x: torch.Tensor,
        carry: dict[str, torch.Tensor],
    ) -> tuple[torch.Tensor, dict[str, torch.Tensor]]:
        if x.dim() != 2:
            raise ValueError("Expected input tensor with shape (batch, timesteps).")

        batch, timesteps = x.size()
        device = x.device
        if "position" not in carry:
            raise KeyError("Carry must include a 'position' tensor.")

        positions = carry["position"]
        if positions.size(0) != batch:
            raise ValueError(
                "Position carry batch dimension does not match the provided tokens."
            )
        if positions.device != device:
            raise ValueError(
                "Position tensor must live on the same device as input tokens."
            )
        if torch.any(positions >= self.max_position_embeddings):
            raise ValueError(
                "Position index exceeds configured positional embedding range."
            )

        position_offsets = torch.arange(timesteps, device=device, dtype=positions.dtype)
        position_indices = positions.unsqueeze(1) + position_offsets
        if torch.any(position_indices >= self.max_position_embeddings):
            raise ValueError(
                "Position index exceeds configured positional embedding range."
            )

        hidden = self.token_embedding(x) + self.position_embedding(position_indices)
        hidden = self.embedding_dropout(hidden)

        updated_carry: dict[str, torch.Tensor] = {}

        for idx, layer in enumerate(self.layers):
            key_name = self.key_names[idx]
            value_name = self.value_names[idx]
            if key_name not in carry or value_name not in carry:
                raise KeyError(
                    "Transformer carry is missing key/value tensors for layer" f" {idx}."
                )
            key_carry = carry[key_name]
            value_carry = carry[value_name]
            if key_carry.size(0) != batch or key_carry.size(1) != self.num_heads:
                raise ValueError(
                    "Key carry shape is incompatible with the provided tokens."
                )
            if value_carry.size(0) != batch or value_carry.size(1) != self.num_heads:
                raise ValueError(
                    "Value carry shape is incompatible with the provided tokens."
                )
            if (key_carry.size(-1) != self.head_dim) or (
                value_carry.size(-1) != self.head_dim
            ):
                raise ValueError("Key/value carry head dimension mismatch detected.")
            if key_carry.device != device or value_carry.device != device:
                raise ValueError("Key/value carry tensors must share the input device.")
            hidden, updated_key_carry, updated_value_carry = layer(
                hidden, key_carry, value_carry
            )
            updated_carry[key_name] = updated_key_carry
            updated_carry[value_name] = updated_value_carry

        hidden = self.final_norm(hidden)
        logits = self.output_projection(hidden)

        updated_carry["position"] = positions + timesteps
        return logits, updated_carry

    @property
    def vocab_size(self) -> int:
        return self._vocab_size


def sinusoidal_position_encoding(
    length: int,
    embedding_dim: int,
    base: float = 10000.0,
) -> Tensor:
    """Create 1D sinusoidal positional embeddings.

    Args:
        length: Number of positions to encode. Must be non-negative.
        embedding_dim: Dimensionality of each embedding. Must be positive.
        base: Exponential base used to compute the angular frequencies.

    Returns:
        A ``(length, embedding_dim)`` tensor of sinusoidal encodings.

    Raises:
        ValueError: If ``length`` is negative, ``embedding_dim`` is not positive,
            or ``base`` is not positive.
    """

    assert length >= 0, "length must be non-negative."
    assert embedding_dim > 0, "embedding_dim must be positive."
    assert base > 0, "base must be positive."

    if length == 0:
        return torch.empty(0, embedding_dim)

    positions = torch.arange(length).unsqueeze(1)
    div_input = torch.arange(0, embedding_dim, 2)
    div_term = torch.exp(div_input * (-math.log(base) / embedding_dim))
    embeddings = torch.zeros(length, embedding_dim)
    angles = positions * div_term
    embeddings[:, 0::2] = torch.sin(angles)

    if embedding_dim % 2 == 0:
        embeddings[:, 1::2] = torch.cos(angles)
    else:
        embeddings[:, 1::2] = torch.cos(angles)[:, : embedding_dim // 2]

    return embeddings


class SinusoidalPositionalEmbedding(nn.Module):
    """Sinusoidal positional embeddings for transformer-style models.

    The module caches a precomputed table of embeddings and extends it on demand.
    Forward accepts either a sequence length or explicit position indices.
    """

    def __init__(
        self,
        embedding_dim: int,
        max_length: int = 2048,
        base: float = 10000.0,
    ) -> None:
        super().__init__()
        assert max_length >= 0, "max_length must be non-negative."
        assert embedding_dim > 0, "embedding_dim must be positive."
        assert base > 0, "base must be positive."

        self.embedding_dim = int(embedding_dim)
        self.base = float(base)

        pe = sinusoidal_position_encoding(max_length, self.embedding_dim, base=self.base)
        self._pe: Tensor
        self.register_buffer("_pe", pe)

    @property
    def pe(self) -> Tensor:
        """Return the cached positional embedding table."""
        return self._pe

    def forward(
        self,
        positions: Optional[Tensor] = None,
        seq_len: Optional[int] = None,
    ) -> Tensor:
        """Look up positional embeddings.

        Args:
            positions: Optional tensor of position indices. Can have any shape,
                and the returned embeddings will append ``embedding_dim`` to that
                shape. Defaults to ``None``.
            seq_len: Optional sequence length. When provided, returns the first
                ``seq_len`` embeddings from the table.

        Returns:
            Tensor of positional embeddings on the same device/dtype as the
            cached table.

        Raises:
            ValueError: If both or neither of ``positions`` and ``seq_len`` are
                provided, or if indices exceed the cached range.
        """

        if (positions is None) == (seq_len is None):
            raise ValueError("Provide exactly one of positions or seq_len.")

        if positions is not None:
            flat_positions = positions.reshape(-1)
            gathered = self._pe.index_select(0, flat_positions)
            return gathered.view(
                positions.shape[0], positions.shape[1], self.embedding_dim
            )
        else:
            return self._pe[:seq_len]


class DiffusionPISTimeEncoding(nn.Module):
    """Time Encoding Module for DiffusionPISGradNet.

    See DiffusionPISGradNet for more details.
    """

    def __init__(self, harmonics_dim: int, t_emb_dim: int, hidden_dim: int) -> None:
        super().__init__()
        self.timestep_phase = nn.Parameter(torch.randn(harmonics_dim)[None])
        self.t_model = nn.Sequential(
            nn.Linear(2 * harmonics_dim, hidden_dim),
            nn.GELU(),
            nn.Linear(hidden_dim, t_emb_dim),
        )
        self.register_buffer(
            # Frequency range [0.1, 100] spans slow-varying to fast-varying
            # harmonics, following the PIS/DIS time-encoding convention
            # (Zhang & Chen 2021, arXiv:2111.15141).
            "pe",
            torch.linspace(start=0.1, end=100, steps=harmonics_dim)[None],
        )

    def forward(self, t: torch.Tensor) -> torch.Tensor:
        """
        Arguments:
            t: torch.Tensor
        """
        t_sin = ((t.unsqueeze(1) * self.pe) + self.timestep_phase).sin()  # type: ignore
        t_cos = ((t.unsqueeze(1) * self.pe) + self.timestep_phase).cos()  # type: ignore
        t_emb = torch.cat([t_sin, t_cos], dim=-1)
        return self.t_model(t_emb)


class DiffusionPISStateEncoding(nn.Module):
    """State Encoding Module for DiffusionPISGradNet.

    See DiffusionPISGradNet for more details.
    """

    def __init__(self, x_dim: int, s_emb_dim: int) -> None:
        super().__init__()

        self.s_model = nn.Linear(x_dim, s_emb_dim)

    def forward(self, s: torch.Tensor) -> torch.Tensor:
        return self.s_model(s)


class DiffusionPISJointPolicy(nn.Module):
    """Joint Policy Module for DiffusionPISGradNet.

    See DiffusionPISGradNet for more details.
    """

    def __init__(
        self,
        s_emb_dim: int,
        hidden_dim: int,
        out_dim: int,
        num_layers: int,
        zero_init: bool = False,
    ) -> None:
        super().__init__()

        self.model = nn.Sequential(
            nn.GELU(),  # Because this model accepts embeddings (linear projections).
            nn.Linear(s_emb_dim, hidden_dim),
            nn.GELU(),
            *[
                nn.Sequential(nn.Linear(hidden_dim, hidden_dim), nn.GELU())
                for _ in range(num_layers - 1)
            ],
            nn.Linear(hidden_dim, out_dim),
        )

        if zero_init:
            self.model[-1].weight.data.fill_(1e-8)  # type: ignore
            self.model[-1].bias.data.fill_(0.0)  # type: ignore

    def forward(self, s_emb: torch.Tensor, t_emb: torch.Tensor) -> torch.Tensor:
        return self.model(s_emb + t_emb)


class DiffusionPISGradNetForward(nn.Module):
    """PISGradNet for diffusion sampling.

    This architecture was first introduced in Path Integral Sampler (PIS) (https://arxiv.org/abs/2111.15141)
    and adapted for GFlowNet-based training by Sendera et al. (https://arxiv.org/abs/2508.03044).

    Attributes:
        s_dim: The dimension of the states.
        harmonics_dim: The dimension of the Fourier features.
        t_emb_dim: The dimension of the time embedding.
        s_emb_dim: The dimension of the state embedding.
        hidden_dim: The dimension of the hidden layers.
        joint_layers: The number of layers in the joint policy.
        zero_init: Whether to initialize the weights and biases of the final layer to zero.
        out_dim: The dimension of the output.
        t_model: The time encoding module.
        s_model: The state encoding module.
        joint_model: The joint policy module.
    """

    def __init__(
        self,
        s_dim: int,  # dimension of states (== target.dim)
        harmonics_dim: int = 64,
        t_emb_dim: int = 64,
        s_emb_dim: int = 64,
        hidden_dim: int = 64,
        joint_layers: int = 2,
        zero_init: bool = False,
        clipping: bool = False,
        gfn_clip: float = 1e4,
        t_scale: float = 1.0,
        log_var_range: float = 4.0,  # kept for parity with learned-var subclass
        learn_variance: bool = False,
        # predict_flow: bool,  # TODO: support predict flow for db or subtb
        # share_embeddings: bool = False,
        # flow_harmonics_dim: int = 64,
        # flow_t_emb_dim: int = 64,
        # flow_s_emb_dim: int = 64,
        # flow_hidden_dim: int = 64,
        # flow_layers: int = 2,
        # lp: bool,  # TODO: support Langevin parameterization
        # lp_layers: int = 3,
        # lp_scaling_per_dimension: bool = True,
        # clipping: bool = False,  # TODO: support clipping
        # out_clip: float = 1e4,
        # lp_clip: float = 1e2,
        # log_var_range: float = 4.0,
    ):
        """Initialize the PISGradNetForward.

        Args:
            s_dim: The dimension of the states.
            harmonics_dim: The dimension of the Fourier features.
            t_emb_dim: The dimension of the time embedding.
            s_emb_dim: The dimension of the state embedding.
            hidden_dim: The dimension of the hidden layers.
            joint_layers: The number of layers in the joint policy.
            zero_init: Whether to initialize the weights and biases of the final layer to zero.
        """
        super().__init__()
        self.s_dim = s_dim
        self.input_dim = s_dim + 1  # + 1 for time, for the default IdentityPreprocessor
        self.harmonics_dim = harmonics_dim
        self.t_emb_dim = t_emb_dim
        self.s_emb_dim = s_emb_dim
        self.hidden_dim = hidden_dim
        self.joint_layers = joint_layers
        self.zero_init = zero_init
        self.learn_variance = learn_variance
        self.out_dim = s_dim + 1 if self.learn_variance else s_dim
        self.clipping = clipping
        self.gfn_clip = gfn_clip
        self.t_scale = t_scale
        self.log_var_range = log_var_range

        assert (
            self.s_emb_dim == self.t_emb_dim
        ), "Dimensionality of state embedding and time embedding should be the same!"

        self.t_model = DiffusionPISTimeEncoding(
            self.harmonics_dim, self.t_emb_dim, self.hidden_dim
        )
        self.s_model = DiffusionPISStateEncoding(self.s_dim, self.s_emb_dim)
        self.joint_model = DiffusionPISJointPolicy(
            self.s_emb_dim,
            self.hidden_dim,
            self.out_dim,
            self.joint_layers,
            self.zero_init,
        )

    def forward(
        self,
        preprocessed_states: torch.Tensor,
        # grad_logr_fn: Callable,  # TODO: grad_logr_fn for lp
    ) -> torch.Tensor:
        """Forward pass of the module.

        Args:
            preprocessed_states: The preprocessed states (shape: (*batch_shape, s_dim + 1))

        Returns:
            The output of the module (shape: (*batch_shape, s_dim)).
        """
        s = preprocessed_states[..., :-1]
        t = preprocessed_states[..., -1]
        s_emb = self.s_model(s)
        t_emb = self.t_model(t)
        out = self.joint_model(s_emb, t_emb)

        if self.learn_variance:
            drift, raw_log_std = out[..., :-1], out[..., [-1]]
            if self.clipping:
                drift = torch.clamp(drift, -self.gfn_clip, self.gfn_clip)
            log_std = torch.tanh(raw_log_std) * self.log_var_range
            out = torch.cat([drift, log_std], dim=-1)
        else:
            if self.clipping:
                out = torch.clamp(out, -self.gfn_clip, self.gfn_clip)

        if torch.isnan(out).any():
            raise ValueError("DiffusionPISGradNetForward produced NaNs")

        return out


class DiffusionFixedBackwardModule(nn.Module):
    """Fixed Backward Module for DiffusionPISGradNet.

    Attributes:
        input_dim: The dimension of the input.
    """

    def __init__(self, s_dim: int):
        """Initialize the FixedBackwardModule.

        Args:
            s_dim: The dimension of the states.
        """
        super().__init__()
        self.input_dim = s_dim + 1  # + 1 for time, for the default IdentityPreprocessor

    def forward(self, preprocessed_states: torch.Tensor) -> torch.Tensor:
        """Forward pass of the module.

        Args:
            preprocessed_states: The preprocessed states (shape: (*batch_shape, s_dim + 1))

        Returns:
            The output of the module (shape: (*batch_shape, s_dim)).
        """
        return torch.zeros_like(preprocessed_states[..., :-1])


class DiffusionPISGradNetBackward(nn.Module):
    """Learnable backward correction module (PIS-style) for diffusion.

    Produces mean and optional log-std corrections that are tanh-scaled by
    `pb_scale_range` to stay close to the analytic Brownian bridge.
    """

    def __init__(
        self,
        s_dim: int,
        harmonics_dim: int = 64,
        t_emb_dim: int = 64,
        s_emb_dim: int = 64,
        hidden_dim: int = 64,
        joint_layers: int = 2,
        zero_init: bool = False,
        clipping: bool = False,
        gfn_clip: float = 1e4,
        pb_scale_range: float = 0.1,
        log_var_range: float = 4.0,
        learn_variance: bool = True,
    ) -> None:
        super().__init__()
        self.s_dim = s_dim
        self.out_dim = s_dim + (1 if learn_variance else 0)
        self.harmonics_dim = harmonics_dim
        self.t_emb_dim = t_emb_dim
        self.s_emb_dim = s_emb_dim
        self.hidden_dim = hidden_dim
        self.joint_layers = joint_layers
        self.zero_init = zero_init
        self.clipping = clipping
        self.gfn_clip = gfn_clip
        self.pb_scale_range = pb_scale_range
        self.log_var_range = log_var_range
        self.learn_variance = learn_variance

        assert (
            self.s_emb_dim == self.t_emb_dim
        ), "Dimensionality of state embedding and time embedding should be the same!"

        self.t_model = DiffusionPISTimeEncoding(
            self.harmonics_dim, self.t_emb_dim, self.hidden_dim
        )
        self.s_model = DiffusionPISStateEncoding(self.s_dim, self.s_emb_dim)
        self.joint_model = DiffusionPISJointPolicy(
            self.s_emb_dim,
            self.hidden_dim,
            self.out_dim,
            self.joint_layers,
            self.zero_init,
        )

    def forward(self, preprocessed_states: torch.Tensor) -> torch.Tensor:
        s = preprocessed_states[..., :-1]
        t = preprocessed_states[..., -1]
        s_emb = self.s_model(s)
        t_emb = self.t_model(t)
        out = self.joint_model(s_emb, t_emb)

        if self.clipping:
            out = torch.clamp(out, -self.gfn_clip, self.gfn_clip)

        # Tanh-bound outputs to [-1, 1]; the estimator
        # (PinnedBrownianMotionBackward.to_probability_distribution) applies the
        # actual pb_scale_range scaling, so we must NOT scale here.
        drift_corr = torch.tanh(out[..., : self.s_dim])
        if self.learn_variance and out.shape[-1] == self.s_dim + 1:
            log_std_corr = torch.tanh(out[..., [-1]])
            log_std_corr = torch.clamp(
                log_std_corr, -self.log_var_range, self.log_var_range
            )
            out = torch.cat([drift_corr, log_std_corr], dim=-1)
        else:
            out = drift_corr

        if torch.isnan(out).any():
            raise ValueError("DiffusionPISGradNetBackward produced NaNs")

        return out



import math
from typing import Callable, Optional

import torch
import torch.distributed as dist

from gfn.containers.message import Message, MessageType
from gfn.containers.replay_buffer import (
    ContainerUnion,
    NormBasedDiversePrioritizedReplayBuffer,
    ReplayBuffer,
)
from gfn.env import Env


class ReplayBufferManager:

    def __init__(
        self,
        env: Env,
        rank: int,
        num_training_ranks: int,
        scoring_function: Optional[Callable[[ContainerUnion], dict[str, float]]] = None,
        diverse_replay_buffer: bool = False,
        capacity: int = 10000,
        remote_manager_rank: int | None = None,
    ):
        self.rank = rank
        self.is_running = True
        self.exit_counter = 0
        self.num_training_ranks = num_training_ranks
        self.scoring_function = scoring_function or self.default_scoring_function
        backend = dist.get_backend()
        if backend != "gloo":
            raise RuntimeError(
                f"Replay Buffer Manager is only supported with the 'gloo' backend, "
                f"but the current backend is '{backend}'."
            )

        self.diverse_replay_buffer = diverse_replay_buffer
        self.capacity = capacity
        self.remote_manager_rank = remote_manager_rank
        if self.diverse_replay_buffer:
            self.replay_buffer = NormBasedDiversePrioritizedReplayBuffer(
                env, capacity=self.capacity
            )
        else:
            self.replay_buffer = ReplayBuffer(
                env,
                capacity=self.capacity,
                prioritized_capacity=True,  # Always prioritize high reward items.
                remote_manager_rank=self.remote_manager_rank,
                remote_buffer_freq=1,
            )

    def default_scoring_function(self, obj) -> dict[str, float]:
        """Default score function if none provided, placeholder."""
        return {"score": math.inf}

    def _compute_metadata(self) -> dict:
        raise NotImplementedError(
            "_compute_metadata is not implemented for default replay buffer manager"
        )

    def run(self):
        """Runs on remote buffer manager ranks. Waits for training data, computes reward, sends back."""

        while self.is_running:
            # Receive data
            sender_rank, msg, _msg_data_len = self._recv_object()

            # Recieved some data to add to the buffer.
            if msg.message_type == MessageType.DATA:
                self.replay_buffer.add(msg.message_data)
                score_dict = self.scoring_function(msg.message_data)
                message = Message(message_type=MessageType.DATA, message_data=score_dict)
                message_tensor = message.serialize()
                length_message_tensor = torch.IntTensor([len(message_tensor)])
                dist.send(length_message_tensor, dst=sender_rank)
                dist.send(message_tensor, dst=sender_rank)

            elif msg.message_type == MessageType.GET_METADATA:
                metadata = self._compute_metadata()
                msg = Message(message_type=MessageType.DATA, message_data=metadata)
                metadata_tensor = msg.serialize()
                length_metadata_tensor = torch.IntTensor([len(metadata_tensor)])
                dist.send(length_metadata_tensor, dst=sender_rank)
                dist.send(metadata_tensor, dst=sender_rank)

            elif msg.message_type == MessageType.EXIT:
                self.exit_counter = self.exit_counter + 1
                if self.exit_counter == self.num_training_ranks:
                    self.is_running = False
                    print(
                        f"Manager - Replay buffer {self.rank} received exit signals from all training ranks. Exiting."
                    )
            else:
                raise ValueError(
                    f"Manager - Rank {self.rank} received unknown message type: {msg.message_type}"
                )

    def _recv_object(self):
        # Receive the length.
        length_tensor = torch.IntTensor([0])
        sender_rank = dist.recv(length_tensor)
        length = length_tensor.item()

        # Receive the actual serialized data.
        byte_tensor = torch.ByteTensor(length)
        dist.recv(byte_tensor, src=sender_rank)

        msg = Message.deserialize(byte_tensor)
        return sender_rank, msg, length

    @staticmethod
    def send_termination_signal(manager_rank: int):
        """Sends a termination signal to the replay buffer manager."""
        rank = dist.get_rank()
        msg = Message(message_type=MessageType.EXIT, message_data=None)
        msg_bytes = msg.serialize()
        length_tensor = torch.IntTensor([len(msg_bytes)])
        dist.send(length_tensor, dst=manager_rank)
        dist.send(msg_bytes, dst=manager_rank)
        print(
            f"Rank {rank} sent termination signal to replay buffer manager {manager_rank}."
        )

    @staticmethod
    def get_metadata(manager_rank: int) -> dict:
        """Sends a get metadata signal to the replay buffer manager."""
        msg = Message(message_type=MessageType.GET_METADATA, message_data=None)
        msg_bytes = msg.serialize()

        length_tensor = torch.IntTensor([len(msg_bytes)])
        dist.send(length_tensor, dst=manager_rank)

        dist.send(msg_bytes, dst=manager_rank)
        length_metadata_tensor = torch.IntTensor([0])

        dist.recv(length_metadata_tensor, src=manager_rank)
        metadata_tensor = torch.ByteTensor(length_metadata_tensor.item())

        dist.recv(metadata_tensor, src=manager_rank)
        metadata = Message.deserialize(metadata_tensor)
        return metadata.message_data



import itertools
from typing import TYPE_CHECKING

import torch
from matplotlib.axes import Axes

if TYPE_CHECKING:
    from gfn.gym.diffusion_sampling import BaseTarget


def sliced_log_reward(
    x: torch.Tensor, target: "BaseTarget", dims: tuple
) -> torch.Tensor:
    _x = torch.zeros((x.shape[0], target.dim))
    _x[:, dims] = x
    return target.log_reward(_x.to(target.device)).detach().cpu()


def viz_2d_slice(
    ax: Axes,
    target: "BaseTarget",
    dims: tuple,
    samples: torch.Tensor | None,
    plot_border: tuple[float, float, float, float],
    alpha=0.5,
    n_contour_levels=50,
    grid_width_n_points=200,
    log_reward_clamp_min=-10000.0,
    use_log_reward=False,
    max_n_samples: int | None = None,
) -> None:
    x_points_dim1 = torch.linspace(plot_border[0], plot_border[1], grid_width_n_points)
    x_points_dim2 = torch.linspace(plot_border[2], plot_border[3], grid_width_n_points)
    x_points = torch.tensor(list(itertools.product(x_points_dim1, x_points_dim2)))
    log_r_x = sliced_log_reward(x=x_points, target=target, dims=dims)
    log_r_x = torch.clamp_min(log_r_x, log_reward_clamp_min)
    log_r_x = log_r_x.reshape((grid_width_n_points, grid_width_n_points))
    if not use_log_reward:
        log_r_x = torch.exp(log_r_x)

    x_points_dim1 = (
        x_points[:, 0].reshape((grid_width_n_points, grid_width_n_points)).numpy()
    )
    x_points_dim2 = (
        x_points[:, 1].reshape((grid_width_n_points, grid_width_n_points)).numpy()
    )
    ax.contour(x_points_dim1, x_points_dim2, log_r_x, levels=n_contour_levels)

    if samples is not None:
        if max_n_samples is not None:
            samples = samples[:max_n_samples]
        samples = samples[:, dims].detach().cpu()
        samples[:, 0] = torch.clamp(samples[:, 0], plot_border[0], plot_border[1])
        samples[:, 1] = torch.clamp(samples[:, 1], plot_border[2], plot_border[3])
        ax.scatter(samples[:, 0], samples[:, 1], alpha=alpha, c="r", marker="x")



# Code copied from https://github.com/google-research/circuit_training
# @article{mirhoseini2021graph,
#   title={A graph placement methodology for fast chip design},
#   author={Mirhoseini*, Azalia and Goldie*, Anna and Yazgan, Mustafa and Jiang, Joe
#   Wenjie and Songhori, Ebrahim and Wang, Shen and Lee, Young-Joon and Johnson,
#   Eric and Pathak, Omkar and Nazi, Azade and Pak, Jiwoo and Tong, Andy and
#   Srinivasa, Kavya and Hang, William and Tuncer, Emre and V. Le, Quoc and
#   Laudon, James and Ho, Richard and Carpenter, Roger and Dean, Jeff},
#   journal={Nature},
#   volume={594},
#   number={7862},
#   pages={207--212},
#   year={2021},
#   publisher={Nature Publishing Group}
# }
#
# Requires Linux x86-64. plc_wrapper_main is a closed-source binary from Google.
#
# Setup:
#   Linux x86-64 (native):
#     cd src/gfn/gym/helpers/chip_design && ./setup_plc.sh
#
#   HPC (via Apptainer/Singularity):
#     cd src/gfn/gym/helpers/chip_design && ./setup_plc.sh --container-only
#
#   Explicit path:
#     export PLC_WRAPPER_MAIN=/path/to/plc_wrapper_main

import os

SAMPLE_NETLIST_FILE = os.path.join(
    os.path.dirname(__file__), "test_data", "netlist.pb.txt"
)
SAMPLE_INIT_PLACEMENT = os.path.join(
    os.path.dirname(__file__), "test_data", "initial.plc"
)
DEFAULT_SIF_IMAGE = os.path.join(os.path.dirname(__file__), "plc_wrapper.sif")
SINGULARITY_DEF = os.path.join(os.path.dirname(__file__), "plc_wrapper.def")

MEDIUM_NETLIST_FILE = os.path.join(
    os.path.dirname(__file__), "test_data", "netlist_medium.pb.txt"
)
MEDIUM_INIT_PLACEMENT = os.path.join(
    os.path.dirname(__file__), "test_data", "initial_medium.plc"
)

PLC_DOWNLOAD_URL = (
    "https://storage.googleapis.com/rl-infra-public/"
    "circuit-training/placement_cost/plc_wrapper_main_{version}"
)



"""GFlowNet environment for chip placement."""

from dataclasses import dataclass
from typing import ClassVar, Dict, List, Optional, Sequence, Tuple, cast

import torch

from gfn.actions import Actions
from gfn.env import DiscreteEnv
from gfn.gym.helpers.chip_design import (
    SAMPLE_INIT_PLACEMENT,
    SAMPLE_NETLIST_FILE,
)
from gfn.gym.helpers.chip_design import utils as placement_util
from gfn.gym.helpers.chip_design.utils import cost_info_function, optimize_orientations
from gfn.states import DiscreteStates


@dataclass
class CostStats:
    """Pre-computed cost statistics for a netlist."""

    mean: float
    std: float
    min: float
    range: float  # max - min

    @staticmethod
    def precompute(
        netlists: List[Tuple[str, str]],
        n_samples: int = 100,
        wirelength_weight: float = 1.0,
        density_weight: float = 1.0,
        congestion_weight: float = 0.5,
        singularity_image: Optional[str] = None,
    ) -> Dict[str, "CostStats"]:
        """Pre-compute cost statistics for multiple netlists.

        Args:
            netlists: List of ``(netlist_file, init_placement)`` tuples.
            n_samples: Number of random placements per netlist.
            wirelength_weight: Wirelength weight for cost computation.
            density_weight: Density weight for cost computation.
            congestion_weight: Congestion weight for cost computation.
            singularity_image: Optional singularity image path.

        Returns:
            Dict mapping netlist file paths to their ``CostStats``.
        """
        stats = {}
        for netlist_file, init_placement in netlists:
            env = ChipDesign(
                netlist_file=netlist_file,
                init_placement=init_placement,
                wirelength_weight=wirelength_weight,
                density_weight=density_weight,
                congestion_weight=congestion_weight,
                singularity_image=singularity_image,
                cd_finetune=False,
                reward_norm=None,
            )
            env._estimate_cost_stats(n_samples)
            stats[netlist_file] = CostStats(
                mean=env._cost_mean,
                std=env._cost_std,
                min=env._cost_min,
                range=env._cost_max,
            )
            env.close()
        return stats


class ChipDesignStates(DiscreteStates):
    """A class to represent the states of the chip design environment."""

    state_shape: ClassVar[tuple[int, ...]]
    s0: ClassVar[torch.Tensor]
    sf: ClassVar[torch.Tensor]
    n_actions: ClassVar[int]

    def __init__(
        self,
        tensor: torch.Tensor,
        current_node_idx: Optional[torch.Tensor] = None,
        conditions: Optional[torch.Tensor] = None,
        device: Optional[torch.device] = None,
        debug: bool = False,
    ):
        super().__init__(
            tensor=tensor, conditions=conditions, device=device, debug=debug
        )
        if current_node_idx is None:
            is_unplaced = tensor == -1
            is_unplaced_padded = torch.cat(
                [
                    is_unplaced,
                    torch.ones_like(is_unplaced[..., :1]),
                ],
                dim=-1,
            )
            current_node_idx = is_unplaced_padded.long().argmax(dim=-1)

        self.current_node_idx = current_node_idx

    def clone(self) -> "ChipDesignStates":
        """Creates a copy of the states."""
        cloned = self.__class__(
            self.tensor.clone(),
            current_node_idx=self.current_node_idx.clone(),
            conditions=(
                self.conditions.clone() if self.conditions is not None else None
            ),
            device=self.tensor.device,
            debug=self.debug,
        )
        if self._forward_masks_cache is not None:
            cloned.forward_masks = self._forward_masks_cache.clone()
        if self._backward_masks_cache is not None:
            cloned.backward_masks = self._backward_masks_cache.clone()
        return cloned

    def __getitem__(self, index) -> "ChipDesignStates":
        """Gets a subset of the states."""
        subset = self.__class__(
            self.tensor[index],
            current_node_idx=self.current_node_idx[index],
        )
        if self._forward_masks_cache is not None:
            subset.forward_masks = self._forward_masks_cache[index]
        if self._backward_masks_cache is not None:
            subset.backward_masks = self._backward_masks_cache[index]
        return subset

    def __setitem__(self, index, value: "ChipDesignStates") -> None:
        """Sets a subset of the states."""
        super().__setitem__(index, value)
        self.current_node_idx[index] = value.current_node_idx

    def extend(self, other: "ChipDesignStates") -> None:
        """Extends the states with another states."""
        super().extend(other)
        self.current_node_idx = torch.cat(
            (self.current_node_idx, other.current_node_idx),
            dim=len(self.batch_shape) - 1,
        )

    @classmethod
    def stack(cls, states: Sequence["ChipDesignStates"]) -> "ChipDesignStates":
        """Stacks the states with another states."""
        stacked = super().stack(states)
        stacked.current_node_idx = torch.stack(
            [s.current_node_idx for s in states],
            dim=0,
        )
        return cast(ChipDesignStates, stacked)


class ChipDesign(DiscreteEnv):
    """
    GFlowNet environment for chip placement.

    The state is a vector of length `n_macros`, where `state[i]` is the grid
    cell location of the i-th macro to be placed. Unplaced macros have a
    location of -1.

    Actions are integers from `0` to `n_grid_cells - 1`, representing the
    grid cell to place the current macro on. Action `n_grid_cells` is the
    exit action.
    """

    def __init__(
        self,
        netlist_file: str = SAMPLE_NETLIST_FILE,
        init_placement: str = SAMPLE_INIT_PLACEMENT,
        std_cell_placer_mode: str = "fd",
        wirelength_weight: float = 1.0,
        density_weight: float = 1.0,
        congestion_weight: float = 0.5,
        device: str | None = None,
        debug: bool = False,
        singularity_image: Optional[str] = None,
        cd_finetune: bool = True,
        reward_norm: Optional[str] = None,
        reward_temper: float = 1.0,
        reward_norm_samples: int = 100,
        cost_stats: Optional[CostStats] = None,
    ):
        """Initializes the chip design environment.

        Args:
            reward_norm: Reward normalization mode. ``None`` for raw
                ``-cost``, ``"zscore"`` for z-score normalization, or
                ``"minmax"`` for min-max normalization. Statistics are
                estimated from ``reward_norm_samples`` random placements
                unless ``cost_stats`` is provided.
            reward_temper: Temperature scaling applied after normalization.
                Higher values sharpen the reward distribution.
            reward_norm_samples: Number of random placements used to
                estimate normalization statistics.
            cost_stats: Pre-computed cost statistics. Use
                ``CostStats.precompute()`` to compute these once for
                multiple netlists, then pass them here to avoid
                recomputation.
        """
        if device is None:
            device = "cuda" if torch.cuda.is_available() else "cpu"

        self.plc = placement_util.create_placement_cost(
            netlist_file=netlist_file,
            init_placement=init_placement,
            singularity_image=singularity_image,
        )
        self.std_cell_placer_mode = std_cell_placer_mode
        self.cd_finetune = cd_finetune
        self.reward_norm = reward_norm
        self.reward_temper = reward_temper

        self.wirelength_weight = wirelength_weight
        self.density_weight = density_weight
        self.congestion_weight = congestion_weight

        self._grid_cols, self._grid_rows = self.plc.get_grid_num_columns_rows()
        self.n_grid_cells = self._grid_cols * self._grid_rows

        self._sorted_node_indices = placement_util.get_ordered_node_indices(
            mode="descending_size_macro_first", plc=self.plc
        )
        self._hard_macro_indices = [
            m for m in self._sorted_node_indices if not self.plc.is_node_soft_macro(m)
        ]
        self.n_macros = len(self._hard_macro_indices)

        # Cost statistics for reward normalization.
        if cost_stats is not None:
            self._cost_mean = cost_stats.mean
            self._cost_std = cost_stats.std
            self._cost_min = cost_stats.min
            self._cost_max = cost_stats.range
        elif reward_norm is not None:
            self._cost_mean = 0.0
            self._cost_std = 1.0
            self._cost_min = 0.0
            self._cost_max = 1.0
            self._estimate_cost_stats(reward_norm_samples)
        else:
            self._cost_mean = 0.0
            self._cost_std = 1.0
            self._cost_min = 0.0
            self._cost_max = 1.0

        s0 = torch.full((self.n_macros,), -1, dtype=torch.long, device=device)
        sf = torch.full((self.n_macros,), -2, dtype=torch.long, device=device)
        n_actions = self.n_grid_cells + 1

        super().__init__(
            n_actions=n_actions,
            s0=s0,
            state_shape=(self.n_macros,),
            sf=sf,
            debug=debug,
        )
        self.States: type[ChipDesignStates] = self.make_states_class()

    def make_states_class(self) -> type[ChipDesignStates]:
        """Creates the ChipDesignStates class."""
        env = self

        class BaseChipDesignStates(ChipDesignStates):
            state_shape = env.state_shape
            s0 = env.s0
            sf = env.sf
            n_actions = env.n_actions

        return BaseChipDesignStates

    def close(self) -> None:
        """Closes the PlacementCost subprocess to free resources."""
        if hasattr(self, "plc") and self.plc is not None:
            self.plc.close()

    def __del__(self) -> None:
        self.close()

    def _apply_state_to_plc(self, state_tensor: torch.Tensor):
        """Applies a single state tensor to the plc object."""
        assert state_tensor.shape == (self.n_macros,)

        self.plc.unplace_all_nodes()
        for i in range(self.n_macros):
            loc = state_tensor[i].item()
            if loc >= 0:
                node_index = self._hard_macro_indices[i]
                self.plc.place_node(node_index, loc)

    def update_masks(self, states: ChipDesignStates) -> None:
        """Updates the forward and backward masks of the states."""
        batch_shape = states.batch_shape
        n = states.tensor.shape[:-1].numel()  # flat batch size

        if n == 0:
            states.forward_masks = torch.zeros(
                *batch_shape, self.n_actions, dtype=torch.bool, device=self.device
            )
            states.backward_masks = torch.zeros(
                *batch_shape, self.n_actions - 1, dtype=torch.bool, device=self.device
            )
            return

        flat_tensor = states.tensor.reshape(n, -1)
        flat_node_idx = states.current_node_idx.reshape(n)
        flat_is_sink = states.is_sink_state.reshape(n)

        fwd = torch.zeros(n, self.n_actions, dtype=torch.bool, device=self.device)
        bwd = torch.zeros(n, self.n_actions - 1, dtype=torch.bool, device=self.device)

        for i in range(n):
            if flat_is_sink[i]:
                continue

            state_tensor = flat_tensor[i]
            current_node_idx = flat_node_idx[i].item()

            if current_node_idx >= self.n_macros:  # All macros placed
                fwd[i, -1] = True  # Only exit is possible
            else:
                self._apply_state_to_plc(state_tensor)
                node_to_place = self._hard_macro_indices[int(current_node_idx)]
                mask = self.plc.get_node_mask(node_to_place)
                mask = torch.tensor(mask, dtype=torch.bool, device=self.device)
                fwd[i, : self.n_grid_cells] = mask

            if current_node_idx > 0:
                last_placed_loc = state_tensor[int(current_node_idx - 1)].item()
                assert last_placed_loc >= 0, "Last placed location should be >= 0"
                bwd[i, int(last_placed_loc)] = True

        states.forward_masks = fwd.reshape(*batch_shape, self.n_actions)
        states.backward_masks = bwd.reshape(*batch_shape, self.n_actions - 1)

    def reset(
        self,
        batch_shape: int | Tuple[int, ...],
        random: bool = False,
        sink: bool = False,
        seed: Optional[int] = None,
        conditions: Optional[torch.Tensor] = None,
    ) -> ChipDesignStates:
        """Resets the environment and computes initial masks."""
        states = super().reset(batch_shape, random, sink, seed, conditions=conditions)
        states = cast(ChipDesignStates, states)
        self.update_masks(states)
        return states

    def _step(self, states: DiscreteStates, actions: Actions) -> ChipDesignStates:
        """Wraps parent _step and updates masks."""
        new_states = super()._step(states, actions)
        new_states = cast(ChipDesignStates, new_states)
        self.update_masks(new_states)
        return new_states

    def _backward_step(
        self, states: DiscreteStates, actions: Actions
    ) -> ChipDesignStates:
        """Wraps parent _backward_step and updates masks."""
        new_states = super()._backward_step(states, actions)
        new_states = cast(ChipDesignStates, new_states)
        self.update_masks(new_states)
        return new_states

    def step(self, states: ChipDesignStates, actions: Actions) -> ChipDesignStates:
        """Performs a forward step in the environment."""
        batch_shape = states.batch_shape
        n = states.tensor.shape[:-1].numel()

        if n == 0:
            return self.States(
                tensor=states.tensor.clone(),
                current_node_idx=states.current_node_idx.clone(),
            )

        flat_tensor = states.tensor.reshape(n, self.n_macros).clone()
        flat_node_idx = states.current_node_idx.reshape(n).clone()
        flat_actions = actions.tensor.reshape(n, actions.tensor.shape[-1])
        flat_is_exit = actions.is_exit.reshape(n)

        non_exit = ~flat_is_exit
        if torch.any(non_exit):
            rows = torch.arange(n, device=self.device)[non_exit]
            cols = flat_node_idx[non_exit]
            flat_tensor[rows, cols] = flat_actions[non_exit].squeeze(-1)

        if torch.any(flat_is_exit):
            flat_tensor[flat_is_exit] = self.sf

        flat_node_idx[non_exit] += 1

        return self.States(
            tensor=flat_tensor.reshape(*batch_shape, self.n_macros),
            current_node_idx=flat_node_idx.reshape(*batch_shape),
        )

    def backward_step(
        self, states: ChipDesignStates, actions: Actions
    ) -> ChipDesignStates:
        """Performs a backward step in the environment."""
        batch_shape = states.batch_shape
        n = states.tensor.shape[:-1].numel()

        flat_tensor = states.tensor.reshape(n, self.n_macros).clone()
        flat_node_idx = states.current_node_idx.reshape(n)

        rows = torch.arange(n, device=self.device)
        cols = flat_node_idx - 1
        flat_tensor[rows, cols] = -1

        new_node_idx = flat_node_idx - 1
        return self.States(
            tensor=flat_tensor.reshape(*batch_shape, self.n_macros),
            current_node_idx=new_node_idx.reshape(*batch_shape),
        )

    def _estimate_cost_stats(self, n_samples: int) -> None:
        """Estimates cost distribution from random placements."""
        import random as pyrandom

        costs = []
        for _ in range(n_samples):
            self.plc.unplace_all_nodes()
            grid_cells = list(range(self.n_grid_cells))
            for macro_idx in self._hard_macro_indices:
                pyrandom.shuffle(grid_cells)
                for cell in grid_cells:
                    if self.plc.can_place_node(macro_idx, cell):
                        self.plc.place_node(macro_idx, cell)
                        break

            self.analytical_placer()
            cost, _ = cost_info_function(
                plc=self.plc,
                done=True,
                wirelength_weight=self.wirelength_weight,
                density_weight=self.density_weight,
                congestion_weight=self.congestion_weight,
            )
            costs.append(cost)

        self.plc.unplace_all_nodes()

        costs_t = torch.tensor(costs, dtype=torch.float64)
        self._cost_mean = costs_t.mean().item()
        self._cost_std = max(costs_t.std().item(), 1e-8)
        self._cost_min = costs_t.min().item()
        self._cost_max = max(costs_t.max().item() - costs_t.min().item(), 1e-8)

    def analytical_placer(self):
        """Places standard cells using an analytical placer."""
        if self.std_cell_placer_mode == "fd":
            placement_util.fd_placement_schedule(self.plc)
        else:
            raise ValueError(
                f"{self.std_cell_placer_mode} is not a supported std_cell_placer_mode."
            )

    def _normalize_cost(self, cost: float) -> float:
        """Applies reward normalization to a raw cost value."""
        if self.reward_norm == "zscore":
            return (cost - self._cost_mean) / self._cost_std
        elif self.reward_norm == "minmax":
            return (cost - self._cost_min) / self._cost_max
        return cost

    def log_reward(self, final_states: ChipDesignStates) -> torch.Tensor:
        """Computes the log reward of the final states."""
        rewards = torch.zeros(len(final_states), device=self.device)
        for i in range(len(final_states)):
            state_tensor = final_states.tensor[i]
            self._apply_state_to_plc(state_tensor)

            self.analytical_placer()

            if self.cd_finetune:
                optimize_orientations(
                    self.plc,
                    wirelength_weight=self.wirelength_weight,
                    density_weight=self.density_weight,
                    congestion_weight=self.congestion_weight,
                )

            cost, _ = cost_info_function(
                plc=self.plc,
                done=True,
                wirelength_weight=self.wirelength_weight,
                density_weight=self.density_weight,
                congestion_weight=self.congestion_weight,
            )
            rewards[i] = -self._normalize_cost(cost) * self.reward_temper
        return rewards



# coding=utf-8
# Copyright 2021 The Circuit Training Team Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
# pyright: reportCallIssue=false, reportArgumentType=false

"""A collection of non-prod utility functions for placement.

All the dependencies in this files should be non-prod.
"""

import datetime
import logging
import re
import textwrap
from typing import Dict, Iterator, List, Optional, Tuple, Union

import numpy as np

from . import plc_client

COST_COMPONENTS = ["wirelength", "congestion", "density"]
ORIENTATIONS = ["N", "S", "E", "W", "FN", "FS", "FE", "FW"]


def optimize_orientations(
    plc: plc_client.PlacementCost,
    wirelength_weight: float = 1.0,
    density_weight: float = 1.0,
    congestion_weight: float = 0.5,
    max_iterations: int = 2,
) -> None:
    """Coordinate descent over macro orientations.

    For each macro, tries all 8 orientations and greedily picks the one
    that minimizes the proxy cost. Repeats for ``max_iterations`` sweeps
    (convergence is typically reached in 1-2).
    """
    macro_indices = list(nodes_of_types(plc, ["MACRO"]))
    if not macro_indices:
        return

    for _ in range(max_iterations):
        improved = False
        for node_index in macro_indices:
            if not plc.is_node_placed(node_index):
                continue
            current_orient = plc.get_macro_orientation(node_index)
            best_orient = current_orient
            best_cost = _proxy_cost(
                plc, wirelength_weight, density_weight, congestion_weight
            )
            for orient in ORIENTATIONS:
                if orient == current_orient:
                    continue
                plc.update_macro_orientation(node_index, orient)
                cost = _proxy_cost(
                    plc, wirelength_weight, density_weight, congestion_weight
                )
                if cost < best_cost:
                    best_cost = cost
                    best_orient = orient
            plc.update_macro_orientation(node_index, best_orient)
            if best_orient != current_orient:
                improved = True
        if not improved:
            break


def _proxy_cost(
    plc: plc_client.PlacementCost,
    wirelength_weight: float,
    density_weight: float,
    congestion_weight: float,
) -> float:
    cost = 0.0
    if wirelength_weight > 0:
        cost += wirelength_weight * plc.get_cost()
    if congestion_weight > 0:
        cost += congestion_weight * plc.get_congestion_cost()
    if density_weight > 0:
        cost += density_weight * plc.get_density_cost()
    return cost


def cost_info_function(
    plc: plc_client.PlacementCost,
    done: bool,
    infeasible_state: bool = False,
    wirelength_weight: float = 1.0,
    density_weight: float = 1.0,
    congestion_weight: float = 0.5,
) -> tuple[float, dict[str, float]]:
    """Returns the RL cost and info."""
    del infeasible_state
    proxy_cost = 0.0
    info = {cost: -1.0 for cost in COST_COMPONENTS}

    if not done:
        return proxy_cost, info

    if wirelength_weight > 0.0:
        info["wirelength"] = plc.get_cost()
        proxy_cost += wirelength_weight * info["wirelength"]

    if congestion_weight > 0.0:
        info["congestion"] = plc.get_congestion_cost()
        proxy_cost += congestion_weight * info["congestion"]

    if density_weight > 0.0:
        info["density"] = plc.get_density_cost()
        proxy_cost += density_weight * info["density"]

    return proxy_cost, info


def nodes_of_types(plc: plc_client.PlacementCost, type_list: List[str]) -> Iterator[int]:
    """Yields the index of a node of certain types."""
    i = 0
    while True:
        node_type = plc.get_node_type(i)
        if not node_type:
            break
        if node_type in type_list:
            yield i
        i += 1


def get_node_xy_coordinates(
    plc: plc_client.PlacementCost,
) -> Dict[int, Tuple[float, float]]:
    """Returns all node x,y coordinates (canvas) in a dict."""
    node_coords = dict()
    for node_index in nodes_of_types(plc, ["MACRO", "STDCELL", "PORT"]):
        if plc.is_node_placed(node_index):
            node_coords[node_index] = plc.get_node_location(node_index)
    return node_coords


def get_macro_orientations(plc: plc_client.PlacementCost) -> Dict[int, int]:
    """Returns all macros' orientations in a dict."""
    macro_orientations = dict()
    for node_index in nodes_of_types(plc, ["MACRO"]):
        macro_orientations[node_index] = plc.get_macro_orientation(node_index)
    return macro_orientations


def restore_node_xy_coordinates(
    plc: plc_client.PlacementCost, node_coords: Dict[int, Tuple[float, float]]
) -> None:
    for node_index, coords in node_coords.items():
        if not plc.is_node_fixed(node_index):
            plc.update_node_coords(node_index, coords[0], coords[1])


def restore_macro_orientations(
    plc: plc_client.PlacementCost, macro_orientations: Dict[int, int]
) -> None:
    for node_index, orientation in macro_orientations.items():
        plc.update_macro_orientation(node_index, orientation)


def extract_attribute_from_comments(
    attribute: str, filenames: List[str]
) -> Optional[str]:
    """Parses the files' comments section, tries to extract the attribute."""
    for filename in filenames:
        if filename:
            f = filename.split(",")[0]
            if f:
                with open(f, "r") as infile:
                    for line in infile:
                        if line.startswith("#"):
                            match = re.search(rf"{attribute} : ([-\w]+)", line)
                            if match:
                                return match.group(1)
                        else:
                            break
    return None


def get_blockages_from_comments(
    filenames: Union[str, List[str]],
) -> Optional[List[List[float]]]:
    """Returns list of blockages if they exist in the file's comments section."""
    if isinstance(filenames, str):
        filenames = [filenames]
    for filename in filenames:
        if not filename:
            continue
        blockages = []
        try:
            with open(filename, "r") as infile:
                for line in infile:
                    if line.startswith("# Blockage : "):
                        blockages.append([float(x) for x in line.split()[3:8]])
                    elif not line.startswith("#"):
                        break
        except OSError:
            logging.error("could not read file %s.", filename)
        if blockages:
            return blockages


def extract_sizes_from_comments(
    filenames: List[str],
) -> Tuple[Optional[float], Optional[float], Optional[int], Optional[int]]:
    """Parses the file's comments section, tries to extract canvas/grid sizes."""
    canvas_width, canvas_height = None, None
    grid_cols, grid_rows = None, None
    for filename in filenames:
        if not filename:
            continue
        with open(filename, "r") as infile:
            for line in infile:
                if line.startswith("#"):
                    fp_re = re.search(
                        r"FP bbox: {([\d\.]+) ([\d\.]+)} {([\d\.]+) ([\d\.]+)}", line
                    )
                    if fp_re:
                        canvas_width = float(fp_re.group(3))
                        canvas_height = float(fp_re.group(4))
                        continue
                    plc_wh = re.search(r"Width : ([\d\.]+)  Height : ([\d\.]+)", line)
                    if plc_wh:
                        canvas_width = float(plc_wh.group(1))
                        canvas_height = float(plc_wh.group(2))
                        continue
                    plc_cr = re.search(r"Columns : ([\d]+)  Rows : ([\d]+)", line)
                    if plc_cr:
                        grid_cols = int(plc_cr.group(1))
                        grid_rows = int(plc_cr.group(2))
                else:
                    break
    return canvas_width, canvas_height, grid_cols, grid_rows


def fix_port_coordinates(plc: plc_client.PlacementCost) -> None:
    """Find all ports and fix their coordinates."""
    for node in nodes_of_types(plc, ["PORT"]):
        plc.fix_node_coord(node)


def create_placement_cost(
    netlist_file: str,
    init_placement: str,
    overlap_threshold: float = 4e-3,
    congestion_smooth_range: int = 5,
    macro_macro_x_spacing: float = 0.1,
    macro_macro_y_spacing: float = 0.1,
    boundary_check: bool = False,
    horizontal_routes_per_micron: float = 70.33,
    vertical_routes_per_micron: float = 74.51,
    macro_horizontal_routing_allocation: float = 51.79,
    macro_vertical_routing_allocation: float = 51.79,
    routes_per_congestion_grid: int = 1000,
    blockages: Optional[List[List[float]]] = None,
    fixed_macro_names_regex: Optional[List[str]] = None,
    legacy_congestion_grid: bool = False,
    singularity_image: Optional[str] = None,
) -> plc_client.PlacementCost:
    """Creates a placement_cost object."""
    if not netlist_file:
        raise ValueError("netlist_file should be provided.")

    block_name = extract_attribute_from_comments("Block", [init_placement, netlist_file])
    if not block_name:
        logging.warning(
            "block_name is not set. Please add the block_name in:\n%s\nor in:\n%s",
            netlist_file,
            init_placement,
        )

    plc = plc_client.PlacementCost(
        netlist_file,
        macro_macro_x_spacing=macro_macro_x_spacing,
        macro_macro_y_spacing=macro_macro_y_spacing,
        singularity_image=singularity_image,
    )

    plc.make_soft_macros_square()

    blockages = blockages or get_blockages_from_comments([netlist_file, init_placement])
    if blockages:
        for blockage in blockages:
            plc.create_blockage(*blockage)

    canvas_width, canvas_height, grid_cols, grid_rows = extract_sizes_from_comments(
        [netlist_file, init_placement]
    )
    if canvas_width and canvas_height:
        plc.set_canvas_size(canvas_width, canvas_height)
    if grid_cols and grid_rows:
        plc.set_placement_grid(grid_cols, grid_rows)
        if legacy_congestion_grid:
            plc.set_congestion_grid(grid_cols, grid_rows)

    plc.set_project_name("circuit_training")
    plc.set_block_name(block_name or "unset_block")
    plc.set_routes_per_micron(horizontal_routes_per_micron, vertical_routes_per_micron)
    plc.set_macro_routing_allocation(
        macro_horizontal_routing_allocation, macro_vertical_routing_allocation
    )
    plc.set_congestion_smooth_range(congestion_smooth_range)

    if not legacy_congestion_grid:
        congestion_grid_size = (
            2.0
            * routes_per_congestion_grid
            / (horizontal_routes_per_micron + vertical_routes_per_micron)
        )
        canvas_width, canvas_height = plc.get_canvas_width_height()
        congestion_grid_cols = max(1, int(canvas_width / congestion_grid_size))
        congestion_grid_rows = max(1, int(canvas_height / congestion_grid_size))
        plc.set_congestion_grid(congestion_grid_cols, congestion_grid_rows)

    plc.set_overlap_threshold(overlap_threshold)
    plc.set_canvas_boundary_check(boundary_check)
    if init_placement:
        plc.restore_placement(init_placement)
        fix_port_coordinates(plc)

    if fixed_macro_names_regex:
        logging.info("Fixing macro locations using regex.")
        fix_macros_by_regex(plc, fixed_macro_names_regex)

    return plc


def get_node_type_counts(plc: plc_client.PlacementCost) -> Dict[str, int]:
    """Returns number of each type of nodes in the netlist."""
    counts = {
        "MACRO": 0,
        "STDCELL": 0,
        "PORT": 0,
        "MACRO_PIN": 0,
        "SOFT_MACRO": 0,
        "HARD_MACRO": 0,
        "SOFT_MACRO_PIN": 0,
        "HARD_MACRO_PIN": 0,
    }

    for node_index in nodes_of_types(plc, ["MACRO", "STDCELL", "PORT", "MACRO_PIN"]):
        node_type = plc.get_node_type(node_index)
        counts[node_type] += 1
        if node_type == "MACRO":
            if plc.is_node_soft_macro(node_index):
                counts["SOFT_MACRO"] += 1
            else:
                counts["HARD_MACRO"] += 1
        if node_type == "MACRO_PIN":
            ref_id = plc.get_ref_node_id(node_index)
            if plc.is_node_soft_macro(ref_id):
                counts["SOFT_MACRO_PIN"] += 1
            else:
                counts["HARD_MACRO_PIN"] += 1
    return counts


def make_blockage_text(plc: plc_client.PlacementCost) -> str:
    ret = ""
    for blockage in plc.get_blockages():
        ret += "Blockage : {}\n".format(" ".join([str(b) for b in blockage]))
    return ret


def save_placement(
    plc: plc_client.PlacementCost, filename: str, user_comments: str = ""
) -> None:
    """Saves the placement file with some information in the comments section."""
    cols, rows = plc.get_grid_num_columns_rows()
    congestion_cols, congestion_rows = plc.get_congestion_grid_num_columns_rows()
    width, height = plc.get_canvas_width_height()
    hor_routes, ver_routes = plc.get_routes_per_micron()
    hor_macro_alloc, ver_macro_alloc = plc.get_macro_routing_allocation()
    smooth = plc.get_congestion_smooth_range()
    info = textwrap.dedent(
        """\
    Placement file for Circuit Training
    Source input file(s) : {src_filename}
    This file : {filename}
    Date : {date}
    Columns : {cols}  Rows : {rows}
    Congestion Columns : {congestion_cols}  Congestion Rows : {congestion_rows}
    Width : {width:.3f}  Height : {height:.3f}
    Area : {area}
    Wirelength : {wl:.3f}
    Wirelength cost : {wlc:.4f}
    Congestion cost : {cong:.4f}
    Density cost : {density:.4f}
    Project : {project}
    Block : {block_name}
    Routes per micron, hor : {hor_routes:.3f}  ver : {ver_routes:.3f}
    Routes used by macros, hor : {hor_macro_alloc:.3f}  ver : {ver_macro_alloc:.3f}
    Smoothing factor : {smooth}
    Overlap threshold : {overlap_threshold}
  """.format(
            src_filename=plc.get_source_filename(),
            filename=filename,
            date=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            cols=cols,
            rows=rows,
            width=width,
            congestion_cols=congestion_cols,
            congestion_rows=congestion_rows,
            height=height,
            area=plc.get_area(),
            wl=plc.get_wirelength(),
            wlc=plc.get_cost(),
            cong=plc.get_congestion_cost(),
            density=plc.get_density_cost(),
            project=plc.get_project_name(),
            block_name=plc.get_block_name(),
            hor_routes=hor_routes,
            ver_routes=ver_routes,
            hor_macro_alloc=hor_macro_alloc,
            ver_macro_alloc=ver_macro_alloc,
            smooth=smooth,
            overlap_threshold=plc.get_overlap_threshold(),
        )
    )

    info += "\n" + make_blockage_text(plc) + "\n"
    info += "\nCounts of node types:\n"
    node_type_counts = get_node_type_counts(plc)
    for node_type in sorted(node_type_counts):
        info += "{:<15} : {:>9}\n".format(node_type + "s", node_type_counts[node_type])
    if user_comments:
        info += "\nUser comments:\n" + user_comments + "\n"
    info += "\nnode_index x y orientation fixed"
    return plc.save_placement(filename, info)


def fd_placement_schedule(
    plc: plc_client.PlacementCost,
    num_steps: Tuple[int, ...] = (100, 100, 100),
    io_factor: float = 1.0,
    move_distance_factors: Tuple[float, ...] = (1.0, 1.0, 1.0),
    attract_factor: Tuple[float, ...] = (100.0, 1.0e-3, 1.0e-5),
    repel_factor: Tuple[float, ...] = (0.0, 1.0e6, 1.0e7),
    use_current_loc: bool = False,
    move_macros: bool = False,
) -> None:
    """A placement schedule that uses force directed method."""
    assert len(num_steps) == len(move_distance_factors)
    assert len(num_steps) == len(repel_factor)
    assert len(num_steps) == len(attract_factor)
    canvas_size = max(plc.get_canvas_width_height())
    max_move_distance = [
        f * canvas_size / s for s, f in zip(num_steps, move_distance_factors)
    ]
    move_stdcells = True
    log_scale_conns = False
    use_sizes = False
    plc.optimize_stdcells(
        use_current_loc,
        move_stdcells,
        move_macros,
        log_scale_conns,
        use_sizes,
        io_factor,
        num_steps,
        max_move_distance,
        attract_factor,
        repel_factor,
    )


def read_node_order_file(
    plc: plc_client.PlacementCost, node_order_file: str
) -> List[int]:
    """Reads the node order from a file."""
    with open(node_order_file, "r") as f:
        node_order = [plc.get_node_index(line.strip()) for line in f.readlines()]
    return node_order


def save_node_order_file(
    plc: plc_client.PlacementCost,
    node_order: List[int],
    node_order_file: str,
) -> None:
    """Saves the node order to a file."""
    with open(node_order_file, "w") as f:
        for node_index in node_order:
            if not plc.is_node_soft_macro(node_index):
                f.write(plc.get_node_name(node_index) + "\n")


def get_ordered_node_indices(
    mode: str,
    plc: plc_client.PlacementCost,
    seed: int = 111,
    node_order_file: str = "",
    exclude_fixed_nodes: bool = True,
) -> List[int]:
    """Returns an ordering of node indices according to the specified mode."""
    rng = np.random.default_rng(seed=seed)
    macro_indices = plc.get_macro_indices()
    hard_macro_indices = [m for m in macro_indices if not plc.is_node_soft_macro(m)]
    soft_macro_indices = [m for m in macro_indices if plc.is_node_soft_macro(m)]

    def macro_area(idx):
        if idx not in hard_macro_indices:
            return 0.0
        w, h = plc.get_node_width_height(idx)
        return w * h

    canvas_width, canvas_height = plc.get_canvas_width_height()

    def distance_to_edge(idx):
        x, y = plc.get_node_location(idx)
        return min(
            x, y, canvas_width - x - canvas_width, canvas_height - y - canvas_height
        )

    logging.info("node_order: %s", mode)
    if mode == "legalization_order":
        ordered_indices = sorted(
            hard_macro_indices,
            key=distance_to_edge,
        ) + sorted(
            soft_macro_indices,
            key=macro_area,
            reverse=True,
        )
    elif mode == "descending_size_macro_first":
        ordered_indices = sorted(
            hard_macro_indices,
            key=macro_area,
            reverse=True,
        ) + sorted(
            soft_macro_indices,
            key=macro_area,
            reverse=True,
        )
    elif mode == "random":
        rng.shuffle(macro_indices)
        ordered_indices = macro_indices
    elif mode == "random_macro_first":
        rng.shuffle(hard_macro_indices)
        logging.info("ordered hard macros: %s", hard_macro_indices)
        ordered_indices = hard_macro_indices + soft_macro_indices
    elif mode == "fake_net_topological":
        fake_net_adj = {}
        fake_nets = plc.get_fake_nets()
        nodes = (
            set([nm[0] for _, nm in fake_nets])
            .union(set([nm[1] for _, nm in fake_nets]))
            .union(set(hard_macro_indices))
        )
        is_port = {n: n not in hard_macro_indices for n in nodes}
        macro_with_fake_net = {n: False for n in nodes}
        for fake_net in fake_nets:
            weight = fake_net[0]
            if weight <= 0:
                continue
            node_0 = fake_net[1][0]
            node_1 = fake_net[1][1]
            fake_net_adj[(node_0, node_1)] = weight
            fake_net_adj[(node_1, node_0)] = weight
            if node_0 in hard_macro_indices:
                macro_with_fake_net[node_0] = True
            if node_1 in hard_macro_indices:
                macro_with_fake_net[node_1] = True

        closeness = {n: 0.0 for n in nodes}

        source = max(nodes, key=lambda n: (is_port[n], macro_area(n)))
        visited_nodes = [source]
        last_node = source
        del closeness[last_node]

        while len(visited_nodes) < len(nodes):
            for node in nodes:
                if node in visited_nodes:
                    continue
                if (node, last_node) in fake_net_adj:
                    closeness[node] += fake_net_adj[(node, last_node)]

            last_node = max(
                closeness,
                key=lambda n: (closeness[n], macro_with_fake_net[n], macro_area(n)),
            )
            visited_nodes.append(last_node)
            del closeness[last_node]

        ordered_indices = [n for n in visited_nodes if n in hard_macro_indices] + sorted(
            soft_macro_indices, key=macro_area
        )[::-1]
    elif mode == "file":
        ordered_indices = read_node_order_file(plc, node_order_file)
    else:
        raise ValueError("{} is an unsupported node placement mode.".format(mode))

    if exclude_fixed_nodes:
        ordered_indices = [m for m in ordered_indices if not plc.is_node_fixed(m)]
    return ordered_indices


def extract_blockages_from_file(
    filename: str, canvas_width: float, canvas_height: float
) -> Optional[List[List[float]]]:
    """Reads blockage information from a given file."""
    blockages = []
    try:
        with open(filename, "r") as infile:
            for line in infile:
                if line.startswith("#"):
                    continue
                items = line.split()
                if len(items) != 4:
                    raise ValueError(
                        "Blockage file does not meet expected format"
                        "Expected format <llx> <lly> <urx> <ury>"
                    )
                llx = float(items[0])
                lly = float(items[1])
                urx = float(items[2])
                ury = float(items[3])
                if llx >= urx:
                    raise ValueError(f"Illegal blockage llx {llx} >= urx {urx}")
                if lly >= ury:
                    raise ValueError(f"Illegal blockage lly {lly} >= ury {ury}")
                if llx < 0:
                    raise ValueError(f"Illegal blockage llx {llx} < 0")
                if urx > canvas_width:
                    raise ValueError(
                        f"Illegal blockage urx {urx} > canvas width {canvas_width}"
                    )
                if lly < 0:
                    raise ValueError(f"Illegal blockage lly {lly} < 0")
                if ury > canvas_height:
                    raise ValueError(
                        f"Illegal blockage ury {ury} > canvas height {canvas_height}"
                    )
                blockages.append([llx, lly, urx, ury, 0.99])
    except IOError:
        logging.error("Could not read file %s", filename)
    return blockages


def get_node_locations(plc: plc_client.PlacementCost) -> Dict[int, int]:
    """Returns all node grid locations (macros and stdcells) in a dict."""
    node_locations = dict()
    for i in nodes_of_types(plc, ["MACRO", "STDCELL"]):
        node_locations[i] = plc.get_grid_cell_of_node(i)
    return node_locations


def get_node_ordering_by_size(plc: plc_client.PlacementCost) -> List[int]:
    """Returns the list of nodes (macros and stdcells) ordered by area."""
    node_areas = dict()
    for i in nodes_of_types(plc, ["MACRO", "STDCELL"]):
        if plc.is_node_fixed(i):
            continue
        w, h = plc.get_node_width_height(i)
        node_areas[i] = w * h
    return sorted(node_areas, key=node_areas.get, reverse=True)


def grid_locations_near(
    plc: plc_client.PlacementCost, start_grid_index: int
) -> Iterator[int]:
    """Yields node indices closest to the start_grid_index."""
    cols, rows = plc.get_grid_num_columns_rows()
    start_col, start_row = start_grid_index % cols, int(start_grid_index / cols)
    for distance in range(cols + rows):
        for row_offset in range(-distance, distance + 1):
            for col_offset in range(-distance, distance + 1):
                if abs(row_offset) + abs(col_offset) != distance:
                    continue
                new_col = start_col + col_offset
                new_row = start_row + row_offset
                if new_col < 0 or new_row < 0 or new_col >= cols or new_row >= rows:
                    continue
                yield int(new_col + new_row * cols)


def place_near(plc: plc_client.PlacementCost, node_index: int, location: int) -> bool:
    """Places a node (legally) closest to the given location."""
    for loc in grid_locations_near(plc, location):
        if plc.can_place_node(node_index, loc):
            plc.place_node(node_index, loc)
            return True
    return False


def disconnect_high_fanout_nets(
    plc: plc_client.PlacementCost, max_allowed_fanouts: int = 500
) -> None:
    high_fanout_nets = []
    for i in nodes_of_types(plc, ["PORT", "STDCELL", "MACRO_PIN"]):
        num_fanouts = len(plc.get_fan_outs_of_node(i))
        if num_fanouts > max_allowed_fanouts:
            print(
                "Disconnecting node: {} with {} fanouts.".format(
                    plc.get_node_name(i), num_fanouts
                )
            )
            high_fanout_nets.append(i)
    plc.disconnect_nets(high_fanout_nets)


def legalize_placement(plc: plc_client.PlacementCost) -> bool:
    """Places the nodes to legal positions snapping to grid cells."""
    fix_port_coordinates(plc)
    node_locations = get_node_locations(plc)
    previous_xy_coords = get_node_xy_coordinates(plc)
    total_macro_displacement = 0
    total_macros = 0
    plc.unplace_all_nodes()
    ordered_nodes = get_node_ordering_by_size(plc)
    for node in ordered_nodes:
        if not place_near(plc, node, node_locations[node]):
            print("Could not place node")
            return False
        if node in previous_xy_coords and not plc.is_node_soft_macro(node):
            x, y = plc.get_node_location(node)
            px, py = previous_xy_coords[node]
            print(
                "x/y displacement: dx = {}, dy = {}, macro: {}".format(
                    x - px, y - py, plc.get_node_name(node)
                )
            )
            total_macro_displacement += abs(x - px) + abs(y - py)
            total_macros += 1
    print(
        "Total macro displacement: {}, avg: {}".format(
            total_macro_displacement, total_macro_displacement / total_macros
        )
    )
    return True


def fix_macros_by_regex(plc: plc_client.PlacementCost, macro_regex_str_list: List[str]):
    """Fix macro locations given a list of macro name regex strings."""
    regexs = []
    for regex_str in macro_regex_str_list:
        regexs.append(re.compile(regex_str))

    hard_macros = []
    for m in plc.get_macro_indices():
        if plc.is_node_soft_macro(m):
            continue
        hard_macros.append(m)

    total = 0
    for m in plc.get_macro_indices():
        if plc.is_node_soft_macro(m):
            continue
        macro_name = plc.get_node_name(m)
        for regex in regexs:
            if regex.fullmatch(macro_name):
                plc.fix_node_coord(m)
                total += 1
                logging.info("Fixed macro: %s", macro_name)
                continue
    logging.info("Total number of fixed macros: %d", total)


def create_blockages_by_spacing_constraints(
    canvas_width: float,
    canvas_height: float,
    macro_boundary_x_spacing: float = 0,
    macro_boundary_y_spacing: float = 0,
    rectilinear_blockages: Optional[List[List[float]]] = None,
) -> List[List[float]]:
    """Create blockages using macro-to-boundary spacing constraints."""
    blockages = []
    blockage_rate = 0.1
    if macro_boundary_x_spacing:
        assert 0 < macro_boundary_x_spacing <= canvas_width
        blockages.append([0, 0, macro_boundary_x_spacing, canvas_height, blockage_rate])
        blockages.append(
            [
                canvas_width - macro_boundary_x_spacing,
                0,
                canvas_width,
                canvas_height,
                blockage_rate,
            ]
        )
    if macro_boundary_y_spacing:
        assert 0 < macro_boundary_y_spacing <= canvas_height
        blockages.append([0, 0, canvas_width, macro_boundary_y_spacing, blockage_rate])
        blockages.append(
            [
                0,
                canvas_height - macro_boundary_y_spacing,
                canvas_width,
                canvas_height,
                blockage_rate,
            ]
        )
    for rectilinear_blockage in rectilinear_blockages or []:
        minx, miny, maxx, maxy, _ = rectilinear_blockage
        if macro_boundary_x_spacing:
            blockages.append(
                [
                    max(minx - macro_boundary_x_spacing, 0),
                    max(miny - macro_boundary_y_spacing, 0),
                    minx,
                    min(maxy + macro_boundary_y_spacing, canvas_height),
                    blockage_rate,
                ]
            )
            blockages.append(
                [
                    maxx,
                    max(miny - macro_boundary_y_spacing, 0),
                    min(maxx + macro_boundary_x_spacing, canvas_width),
                    min(maxy + macro_boundary_y_spacing, canvas_height),
                    blockage_rate,
                ]
            )
        if macro_boundary_y_spacing:
            blockages.append(
                [
                    max(minx - macro_boundary_x_spacing, 0),
                    max(miny - macro_boundary_y_spacing, 0),
                    min(maxx + macro_boundary_x_spacing, canvas_width),
                    miny,
                    blockage_rate,
                ]
            )
            blockages.append(
                [
                    max(minx - macro_boundary_x_spacing, 0),
                    maxy,
                    min(maxx + macro_boundary_x_spacing, canvas_width),
                    min(maxy + macro_boundary_y_spacing, canvas_height),
                    blockage_rate,
                ]
            )
    return blockages
